'''
   PyMFEM example 34p

   See c++ version in the MFEM library for more detail

   Sample runs:  mpirun -np 4 python ex34p.py -o 2
                 mpirun -np 4 python ex34p.py -pa -hex

   Device sample runs:
                 mpirun -np 4 python ex34p.py -o 2 -pa -hex -d cuda
                 mpirun -np 4 python ex34p.py -o 2 -no-pa -d cuda

'''
import mfem.par as mfem
from mfem.par import intArray, doubleArray
import os
from os.path import expanduser, join
import numpy as np
from numpy import sin, cos, array, pi, sqrt, floor

from mpi4py import MPI
num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
smyid = '.{:0>6d}'.format(myid)

pa = False
algebraic_ceed = False
visualization = True

def run(ser_ref_levels=1,
        par_ref_levels=1,
        order=1,
        delta_const=1e-6,
        static_cond=False,
        mixed=True,
        device_config="cpu"):

    cond_attr = mfem.intArray()

    mesh_file = "../data/fichera-mixed.mesh"
    if not mixed or pa:
        mesh_file = "../data/fichera.mesh"

    if mesh_file == "../data/fichera-mixed.mesh":
        submesh_elems = mfem.intArray([0, 2, 3, 4, 9])
    else:
        submesh_elems = mfem.intArray([10, 14, 34, 36, 37, 38, 39])

    sym_plane_attr = mfem.intArray([9, 10, 11, 12, 13, 14, 15, 16, ])

    phi0_attr = mfem.intArray()
    phi1_attr = mfem.intArray()
    jn_zero_attr = mfem.intArray()

    if (mesh_file == "../data/fichera-mixed.mesh" or
            mesh_file == "../data/fichera.mesh"):
        phi0_attr.Append(2)

    if (mesh_file == "../data/fichera-mixed.mesh" or
            mesh_file == "../data/fichera.mesh"):
        phi1_attr.Append(23)

    if (mesh_file == "../data/fichera-mixed.mesh" or
            mesh_file == "../data/fichera.mesh"):
        jn_zero_attr.Append(25)
    for i in range(sym_plane_attr.Size()):
        jn_zero_attr.Append(sym_plane_attr[i])

    # 3. Enable hardware devices such as GPUs, and programming models such as
    #    CUDA, OCCA, RAJA and OpenMP based on command line options.
    device = mfem.Device(device_config)
    device.Print()

    # 4. Read the (serial) mesh from the given mesh file on all processors.  We
    #    can handle triangular, quadrilateral, tetrahedral, hexahedral, surface
    #    and volume meshes with the same code.
    mesh_file = expanduser(
        join(os.path.dirname(__file__), *mesh_file.split("/")))
    mesh = mfem.Mesh(mesh_file, 1, 1)
    dim = mesh.Dimension()
    if not mixed or pa:
        mesh.UniformRefinement()
        if ser_ref_levels > 0:
            ser_ref_levels = ser_ref_levels - 1
        else:
            par_ref_levels = par_ref_levels - 1
    submesh_attr = -1

    if cond_attr.Size() == 0 and submesh_elems.Size() > 0:
        max_attr = mesh.attributes.Max()
        submesh_attr = max_attr + 1

        for i in range(submesh_elems.Size()):
            mesh.SetAttribute(submesh_elems[i], submesh_attr)
        mesh.SetAttributes()
        if cond_attr.Size() == 0:
            cond_attr.Append(submesh_attr)

    # 5. Refine the serial mesh on all processors to increase the resolution. In
    #    this example we do 'ref_levels' of uniform refinement.

    for l in range(ser_ref_levels):
        mesh.UniformRefinement()

    # 6. Define a parallel mesh by a partitioning of the serial mesh. Refine
    #    this mesh further in parallel to increase the resolution. Once the
    #    parallel mesh is defined, the serial mesh can be deleted.
    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
    del mesh
    for l in range(par_ref_levels):
        pmesh.UniformRefinement()

    # 6b. Extract a submesh covering a portion of the domain
    pmesh_cond = mfem.ParSubMesh.CreateFromDomain(pmesh, cond_attr)

    # 7. Define a suitable finite element space on the SubMesh and compute
    #    the current density as an H(div) field.
    fec_cond_rt = mfem.RT_FECollection(order - 1, dim)
    fes_cond_rt = mfem.ParFiniteElementSpace(pmesh_cond, fec_cond_rt)
    j_cond = mfem.ParGridFunction(fes_cond_rt)

    ComputeCurrentDensityOnSubMesh(
        order, phi0_attr, phi1_attr, jn_zero_attr, j_cond)

    # 7a. Save the SubMesh and associated current density in parallel. This
    #     output can be viewed later using GLVis:
    #        "glvis -np <np> -m cond_mesh -g cond_j"
    pmesh_cond.Print("cond_mesh"+smyid, 8)
    j_cond.Save("cond_j"+smyid, 8)

    # 7b. Send the current density, computed on the SubMesh, to a GLVis server.
    if visualization:
        port_sock = mfem.socketstream("localhost", 19916)
        port_sock << "parallel " << num_procs << " " << myid << "\n"
        port_sock.precision(8)
        port_sock << "solution\n" << pmesh_cond << j_cond
        port_sock << "window_title 'Conductor J'"
        port_sock << "window_geometry 400 0 400 350"
        port_sock.flush()

    # 8. Define a parallel finite element space on the full mesh. Here we use
    #    the H(curl) finite elements for the vector potential and H(div) for the
    #    current density.
    fec_nd = mfem.ND_FECollection(order, dim)
    fec_rt = mfem.RT_FECollection(order - 1, dim)
    fespace_nd = mfem.ParFiniteElementSpace(pmesh, fec_nd)
    fespace_rt = mfem.ParFiniteElementSpace(pmesh, fec_rt)

    j_full = mfem.ParGridFunction(fespace_rt)
    j_full.Assign(0.0)
    pmesh_cond.Transfer(j_cond, j_full)

    # 8a. Send the transferred current density to a GLVis server.
    if visualization:
        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock << "parallel " << num_procs << " " << myid << "\n"
        sol_sock.precision(8)
        sol_sock << "solution\n" << pmesh << j_full
        sol_sock << "window_title 'J Full'"
        sol_sock << "window_geometry 400 430 400 350"
        sol_sock.flush()

    # 9. Determine the list of true (i.e. parallel conforming) essential
    #    boundary dofs. In this example, the boundary conditions are defined by
    #    marking all the boundary attributes except for those on a symmetry
    #    plane as essential (Dirichlet) and converting them to a list of true
    #    dofs.
    ess_tdof_list = mfem.intArray()
    ess_bdr = mfem.intArray()
    if pmesh.bdr_attributes.Size():
        ess_bdr.SetSize(pmesh.bdr_attributes.Max())
        ess_bdr.Assign(1)
        for i in range(sym_plane_attr.Size()):
            ess_bdr[sym_plane_attr[i]-1] = 0
        fespace_nd.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

    # 10. Set up the parallel linear form b(.) which corresponds to the
    #     right-hand side of the FEM linear system, which in this case is
    #     (J,W_i) where J is given by the function H(div) field transferred
    #     from the SubMesh and W_i are the basis functions in the finite
    #     element fespace.
    jCoef = mfem.VectorGridFunctionCoefficient(j_full)
    b = mfem.ParLinearForm(fespace_nd)
    b.AddDomainIntegrator(mfem.VectorFEDomainLFIntegrator(jCoef))
    b.Assemble()

    # 11. Define the solution vector x as a parallel finite element grid
    #     function corresponding to fespace. Initialize x to zero.
    x = mfem.ParGridFunction(fespace_nd)
    x.Assign(0.0)

    # 12. Set up the parallel bilinear form corresponding to the EM diffusion
    #     operator curl muinv curl + delta I, by adding the curl-curl and the
    #     mass domain integrators. For standard magnetostatics equations choose
    #     delta << 1. Larger values of delta should make the linear system
    #     easier to solve at the expense of resembling a diffusive quasistatic
    #     magnetic field.  A reasonable balance must be found whenever the mesh
    #     or problem setup is altered.
    muinv = mfem.ConstantCoefficient(1.0)
    delta = mfem.ConstantCoefficient(delta_const)
    a = mfem.ParBilinearForm(fespace_nd)
    if pa:
        a.SetAssemblyLevel(mfem.AssemblyLevel_PARTIAL)
    a.AddDomainIntegrator(mfem.CurlCurlIntegrator(muinv))
    a.AddDomainIntegrator(mfem.VectorFEMassIntegrator(delta))

    # 13. Assemble the parallel bilinear form and the corresponding linear
    #     system, applying any necessary transformations such as: parallel
    #     assembly, eliminating boundary conditions, applying conforming
    #     constraints for non-conforming AMR, static condensation, etc.
    if static_cond:
        a.EnableStaticCondensation()
    a.Assemble()

    A = mfem.OperatorPtr()
    B = mfem.Vector()
    X = mfem.Vector()
    a.FormLinearSystem(ess_tdof_list, x, b, A, X, B)

    # 14. Solve the system AX=B using PCG with an AMS preconditioner.
    if myid == 0:
        print("\n".join(("\nSolving for magnetic vector potential ",
                         "using CG with AMS")))

    if pa:  # Jacobi preconditioning in partial assembly mode
        ams = mfem.MatrixFreeAMS(a, A, fespace_nd, muinv, delta, None, ess_bdr)
        cg = mfem.CGSolver(MPI.COMM_WORLD)
        cg.SetRelTol(1e-12)
        cg.SetMaxIter(1000)
        cg.SetPrintLevel(1)
        cg.SetOperator(*A)
        cg.SetPreconditioner(ams)
        cg.Mult(B, X)

    else:
        AM = A.AsHypreParMatrix()
        if myid == 0:
            print("\nSize of linear system: "+str(AM.GetGlobalNumRows()))

        prec_fespace = a.SCParFESpace() if a.StaticCondensationIsEnabled() else fespace_nd

        ams = mfem.HypreAMS(AM, prec_fespace)
        pcg = mfem.HyprePCG(AM)
        pcg.SetTol(1e-12)
        pcg.SetMaxIter(500)
        pcg.SetPrintLevel(2)
        pcg.SetPreconditioner(ams)
        pcg.Mult(B, X)

    # 15. Recover the parallel grid function corresponding to X. This is the
    #     local finite element solution on each processor.
    a.RecoverFEMSolution(X, b, x)

    # 16. Save the refined mesh and the solution in parallel. This output can
    #     be viewed later using GLVis: "glvis -np <np> -m mesh -g sol".
    pmesh.Print("mesh"+smyid, 8)
    x.Save("sol"+smyid, 8)

    # 17. Send the solution by socket to a GLVis server.
    if visualization:
        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock << "parallel " << num_procs << " " << myid << "\n"
        sol_sock << "solution\n" << pmesh << x
        sol_sock << "window_title 'Vector Potential'"
        sol_sock << "window_geometry 800 0 400 350"
        sol_sock.flush()

    # 18. Compute the magnetic flux as the curl of the solution
    curl = mfem.ParDiscreteLinearOperator(fespace_nd, fespace_rt)
    curl.AddDomainInterpolator(mfem.CurlInterpolator())
    curl.Assemble()
    curl.Finalize()

    dx = mfem.ParGridFunction(fespace_rt)
    curl.Mult(x, dx)

    # 19. Save the curl of the solution in parallel. This output can be viewed
    #     later using GLVis: "glvis -np <np> -m mesh -g dsol".
    dx.Save("dsol"+smyid, 8)

    # 19. Send the curl of the solution by socket to a GLVis server.
    if visualization:
        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock << "parallel " << num_procs << " " << myid << "\n"
        sol_sock.precision(8)
        sol_sock << "solution\n" << pmesh << dx
        sol_sock << "window_title 'Magnetic Flux'"
        sol_sock << "window_geometry 1200 0 400 350"
        sol_sock.flush()


def ComputeCurrentDensityOnSubMesh(order, phi0_attr, phi1_attr, jn_zero_attr,
                                   j_cond):
    # Extract the finite element space and mesh on which j_cond is defined
    fes_cond_rt = j_cond.ParFESpace()
    pmesh_cond = fes_cond_rt.GetParMesh()
    dim = pmesh_cond.Dimension()

    # Define a parallel finite element space on the SubMesh. Here we use the H1
    # finite elements for the electrostatic potential.
    fec_h1 = mfem.H1_FECollection(order, dim)
    fes_cond_h1 = mfem.ParFiniteElementSpace(pmesh_cond, fec_h1)

    # Define the conductivity coefficient and the boundaries associated with the
    # fixed potentials phi0 and phi1 which will drive the current.
    sigmaCoef = mfem.ConstantCoefficient(1.0)
    ess_bdr_phi = mfem.intArray(pmesh_cond.bdr_attributes.Max())
    ess_bdr_j = mfem.intArray(pmesh_cond.bdr_attributes.Max())
    ess_bdr_tdof_phi = mfem.intArray()
    ess_bdr_phi.Assign(0)
    ess_bdr_j.Assign(0)

    for i in range(phi0_attr.Size()):
        ess_bdr_phi[phi0_attr[i]-1] = 1

    for i in range(phi1_attr.Size()):
        ess_bdr_phi[phi1_attr[i]-1] = 1

    for i in range(jn_zero_attr.Size()):
        ess_bdr_j[jn_zero_attr[i]-1] = 1

    fes_cond_h1.GetEssentialTrueDofs(ess_bdr_phi, ess_bdr_tdof_phi)

    # Setup the bilinear form corresponding to -Div(sigma Grad phi)
    a_h1 = mfem.ParBilinearForm(fes_cond_h1)
    a_h1.AddDomainIntegrator(mfem.DiffusionIntegrator(sigmaCoef))
    a_h1.Assemble()

    # Set the r.h.s. to zero
    b_h1 = mfem.ParLinearForm(fes_cond_h1)
    b_h1.Assign(0.0)

    # Setup the boundary conditions on phi
    one = mfem.ConstantCoefficient(1.0)
    zero = mfem.ConstantCoefficient(0.0)
    phi_h1 = mfem.GridFunction(fes_cond_h1)
    phi_h1.Assign(0.0)

    bdr0 = mfem.intArray([0]*pmesh_cond.bdr_attributes.Max())
    for i in range(phi0_attr.Size()):
        bdr0[phi0_attr[i]-1] = 1

    phi_h1.ProjectBdrCoefficient(zero, bdr0)

    bdr1 = mfem.intArray([0]*pmesh_cond.bdr_attributes.Max())
    for i in range(phi1_attr.Size()):
        bdr1[phi1_attr[i]-1] = 1

    phi_h1.ProjectBdrCoefficient(one, bdr1)

    A = mfem.OperatorPtr()
    B = mfem.Vector()
    X = mfem.Vector()
    a_h1.FormLinearSystem(ess_bdr_tdof_phi, phi_h1, b_h1, A, X, B)

    # Solve the linear system using algebraic multigrid
    if myid == 0:
        print("\nSolving for electric potential using CG with AMG")
    A = mfem.OperatorPtr()
    B = mfem.Vector()
    X = mfem.Vector()
    a_h1.FormLinearSystem(ess_bdr_tdof_phi, phi_h1, b_h1, A, X, B)

    prec = mfem.HypreBoomerAMG()
    cg = mfem.CGSolver(MPI.COMM_WORLD)
    cg.SetRelTol(1e-12)
    cg.SetMaxIter(2000)
    cg.SetPrintLevel(1)
    cg.SetPreconditioner(prec)
    cg.SetOperator(A.Ptr())
    cg.Mult(B, X)
    a_h1.RecoverFEMSolution(X, b_h1, phi_h1)

    # visualization
    if visualization:
        port_sock = mfem.socketstream("localhost", 19916)
        port_sock << "parallel " << num_procs << " " << myid << "\n"
        port_sock.precision(8)
        port_sock << "solution\n" << pmesh_cond << phi_h1
        port_sock << "window_title 'Conductor Potential'"
        port_sock << "window_geometry 0 0 400 350"
        port_sock.flush()

    # Solve for the current density J = -sigma Grad phi with boundary conditions
    # J.n = 0 on the walls of the conductor but not on the ports where phi=0 and
    # phi=1.

    # J will be computed in H(div) so we need an RT mass matrix
    m_rt = mfem.ParBilinearForm(fes_cond_rt)
    m_rt.AddDomainIntegrator(mfem.VectorFEMassIntegrator())
    m_rt.Assemble()

    # Assemble the (sigma Grad phi) operator
    d_h1 = mfem.ParMixedBilinearForm(fes_cond_h1, fes_cond_rt)
    d_h1.AddDomainIntegrator(mfem.MixedVectorGradientIntegrator(sigmaCoef))
    d_h1.Assemble()

    # Compute the r.h.s, b_rt = sigma E = -sigma Grad phi
    b_rt = mfem.ParLinearForm(fes_cond_rt)
    d_h1.Mult(phi_h1, b_rt)
    b_rt *= -1.0

    # Apply the necessary boundary conditions and solve for J in H(div)
    glb_size_rt = fes_cond_rt.GlobalTrueVSize()
    if myid == 0:
        print("\nSolving for current density in H(Div), using diagonally scaled CG")
        print("Size of linear system: " + str(glb_size_rt))

    ess_bdr_tdof_rt = mfem.intArray()
    M = mfem.OperatorPtr()
    B = mfem.Vector()
    X = mfem.Vector()

    fes_cond_rt.GetEssentialTrueDofs(ess_bdr_j, ess_bdr_tdof_rt)
    j_cond.Assign(0.0)
    m_rt.FormLinearSystem(ess_bdr_tdof_rt, j_cond, b_rt, M, X, B)

    pred = mfem.HypreDiagScale()
    cg = mfem.CGSolver(MPI.COMM_WORLD)
    cg.SetRelTol(1e-12)
    cg.SetMaxIter(2000)
    cg.SetPrintLevel(1)
    cg.SetPreconditioner(prec)
    cg.SetOperator(M.Ptr())
    cg.Mult(B, X)
    m_rt.RecoverFEMSolution(X, b_rt, j_cond)


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(
        description='Ex34p (Sourse Function using a Submehs Transfer )')

    parser.add_argument('-rs', '--refine-serial',
                        action='store', default=1, type=int,
                        help="Number of times to refine the mesh uniformly in serial.")
    parser.add_argument('-rp', '--refine-parallel',
                        action='store', default=1, type=int,
                        help="Number of times to refine the mesh uniformly in parallel.")
    parser.add_argument("-o", "--order",
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree).")
    parser.add_argument("-mc", "--magnetic-cond",
                        action='store', default=1e-6, type=float,
                        help="Magnetic Conductivity")
    parser.add_argument("-sc", "--static-condensation",
                        action='store_true',
                        help="Enable static condensation.")
    parser.add_argument("-hex", "--hex-mesh",
                        action='store_true',  default=False,
                        help="Mixed mesh of hexahedral mesh.")
    parser.add_argument("-pa", "--partial-assembly",
                        action='store_true',
                        help="Enable Partial Assembly.")
    parser.add_argument("-d", "--device",
                        default="cpu", type=str,
                        help="Device configuration string, see Device::Configure().")
    parser.add_argument("-no-vis", "--no-visualization",
                        action='store_true', default=False,
                        help='Enable GLVis visualization')

    args = parser.parse_args()
    if myid == 0:
        parser.print_options(args)

    globals()["pa"] = args.partial_assembly
    globals()["visualization"] = not args.no_visualization

    run(ser_ref_levels=args.refine_serial,
        par_ref_levels=args.refine_parallel,
        order=args.order,
        delta_const=args.magnetic_cond,
        static_cond=args.static_condensation,
        mixed=(not args.hex_mesh),
        device_config=args.device)



'''
 This example shows
   constructing HYPREParCSR in a distributed manner
     1) build scipy.sparse on each node
           each node has informaiton of small number of rows
     2) assemble it to HYPREParCSR
           scipy matries are concatenated vertical

   convert back ParCSR to scipy.sparse
     1) on each node, it produce scipy.sparse which contain
        only its own rows.


   example_3 tests
      multiplication between square matrix and non-square matrix.
   
      it seems,,
      1) m>=n is requred to generate ParCSR ??
      2) partitioning of diagnal elemnts should be the same,
         which makes sense since diagnals are computed w/o
         communication.



'''
import mfem.par as par
from mfem.common.parcsr_extra import *

from scipy.sparse import csr_matrix, coo_matrix

from mpi4py import MPI
comm     = MPI.COMM_WORLD     
num_proc = MPI.COMM_WORLD.size
myid     = MPI.COMM_WORLD.rank

def print_hypre(M, txt):
    for i in range(num_proc):
       MPI.COMM_WORLD.Barrier()                              
       if myid == i:
          if myid == 0:
              print(txt)
              print('MyID: ', myid)
          else:
              print('MyID: ', myid)
          m  = ToScipyCoo(M)              
          print('shape = ', m.shape)
          print(m.toarray())
       MPI.COMM_WORLD.Barrier()                              
shape = (2, 5) if myid < 2 else (1, 5)
# make sample matrix
row  = np.array([0, 0, 0, 0])
col  = np.array([0, 3, 1, 2])
data = np.array([4, 5, 7, 9])
m = coo_matrix((data, (row, col)), shape=shape)
m = m.tocsr()
m = m*(myid+1)

M = ToHypreParCSR(m)
print_hypre(M, 'matrix M')

# make sample vector, it produce vertical array
M2 = Array2Hypre(np.array([1,2,3,4,5]))
'''
if myid == 0:
   row  = np.array([0, 1, ])
   col  = np.array([0, 0, ])
   data = np.array([4, 10,  ])
   m2 = coo_matrix((data, (row, col)), shape=(2, 3))
   m2 = m2.tocsr()
elif myid == 1:
   row  = np.array([ ])
   col  = np.array([ ])
   data = np.array([ ])
   m2 = coo_matrix((data, (row, col)), shape=(2, 3))
   m2 = m2.tocsr()
#   m2 = None
else:
   row  = np.array([])
   col  = np.array([])
   data = np.array([])
   m2 = coo_matrix((data, (row, col)), shape=(1, 3))
   m2 = m2.tocsr()
M2 = ToHypreParCSR(m2)
'''
#print ToScipyCoo(M2)

# adding matrix
M3 = mfem.ParMult(M, M2)
print_hypre(M3, 'M*v')
v = Hypre2Array(M3)
if myid == 0: print(v)
M3 = mfem.ParMult(M2.Transpose(), M)
print_hypre(M3, 'v^t * M')
v = Hypre2Array(M3)
if myid == 0: print(v)



from __future__ import print_function
import os
import sys
import numpy as np

if len(sys.argv) > 1 and sys.argv[1] == '-p':   
    import mfem.par as mfem
    use_parallel = True
    from mfem.common.mpi_debug import nicePrint as print
    from mpi4py import MPI
    myid  = MPI.COMM_WORLD.rank
    
else:
    import mfem.ser as mfem
    use_parallel = False
    myid = 0
    
def run_test():
    a = mfem.IntegrationPointArray()
    assert a.Size() == 0, "array size is wrong"

    mt = mfem.MemoryType.HOST
    a = mfem.IntegrationPointArray(mt)
    assert a.Size() == 0, "array size is wrong"

    mt = mfem.MemoryType.HOST
    a = mfem.IntegrationPointArray(3, mt)
    assert a.Size() == 3, "array size is wrong"

    #If you do this. be carefull not to rules being GCed.
    #rules = mfem.IntegrationRules()
    #ir = rules.Get(2, 2)    
    ir = mfem.IntRules.Get(2, 2)

    answer = [(0.16666666666666666, 0.16666666666666666),
              (0.16666666666666666, 0.6666666666666667),
              (0.6666666666666667, 0.16666666666666666)]
    
    points = [(ir[i].x, ir[i].y) for i in range(ir.GetNPoints())]
    if points != answer:
        print(points)
        print(answer)
        assert False, "integration points coords are not collect" 
    
    pts = [ir[i] for i in range(ir.GetNPoints())]    
    a = mfem.IntegrationPointArray(pts)
    bbb = mfem.IntegrationPointArray((a.GetData(), 3))

    points2 = [(a[i].x, a[i].y) for i in range(ir.GetNPoints())]

    assert (points2 == points), "IntegrationPointArray coords does not agree (check 1)."

    irs = [mfem.IntRules.Get(i, 2) for i in range(mfem.Geometry.NumGeom)]
    
    rulearray = mfem.IntegrationRuleArray(irs)

    ir2 = rulearray[2]
    points3 = [(ir2[i].x, ir2[i].y) for i in range(ir2.GetNPoints())]
    assert (points3 == points), "IntegrationPointArray coords does not agree (check 2)."

    # check slice of IntegrationRuleArray
    rulearray2 = rulearray[:5]
    assert not rulearray2.OwnsData(), "subarray should not own data"
    ir2 = rulearray2[2]
    points3 = [(ir2[i].x, ir2[i].y) for i in range(ir2.GetNPoints())]
    assert (points3 == points), "IntegrationPointArray coords does not agree (check 3)."

    # create it from a pointer array
    data = rulearray2.GetData()    # this returns <Swig Object of type 'mfem::IntegrationRule **'>
    rulearray3 = mfem.IntegrationRuleArray((data, 5))
    ir2 = rulearray3[2]
    points3 = [(ir2[i].x, ir2[i].y) for i in range(ir2.GetNPoints())]
    assert (points3 == points), "IntegrationPointArray coords does not agree (check 3)."
    
    try:
        a = mfem.IntegrationPointArray([1,2,3])
    except:
        print("this one is supposed to fail")
        
if __name__=='__main__':
    run_test()
    



'''
   MFEM example 15

   How to run:
      python <arguments>

   Example of arguments:
      ex15.py
      ex15.py -o 1 -y 0.4
      ex15.py -o 4 -y 0.1
      ex15.py -n 5
      ex15.py -p 1 -n 3

      Other meshes:

      ex15.py -m square-disc-nurbs.mesh
      ex15.py -m disc-nurbs.mesh
      ex15.py -m fichera.mesh -tf 0.3
      ex15.py -m ball-nurbs.mesh -tf 0.3
      ex15.py -m mobius-strip.mesh
      ex15.py -m amr-quad.mesh

      Conforming meshes (no derefinement):

      ex15.py -m square-disc.mesh
      ex15.py -m escher.mesh -r 2 -tf 0.3

'''

import sys
from mfem.common.arg_parser import ArgParser
from os.path import expanduser, join, dirname
import numpy as np
from numpy import cos, sin, pi, exp, sqrt, arctan

import mfem.ser as mfem
from mfem.ser import intArray

try:
    from numba import jit, types, jit_module
    HAS_NUMBA = True
except ImportError:
    HAS_NUMBA = False


parser = ArgParser(description='Ex15')
parser.add_argument('-m', '--mesh',
                    default='star-hilbert.mesh',
                    action='store', type=str,
                    help='Mesh file to use.')
parser.add_argument("-p", "--problem",
                    action='store', default=0, type=int,
                    help="Problem setup to use: 0 = spherical front, 1 = ball.")
parser.add_argument("-n", "--nfeatures",
                    action='store', default=1, type=int,
                    help="Number of solution features (fronts/balls).")
parser.add_argument('-o', '--order',
                    action='store', default=2, type=int,
                    help="Finite element order (polynomial degree)")
parser.add_argument("-e", "--max-err",
                    action='store', default=5e-3, type=float,
                    help="Maximum element error")
parser.add_argument("-y", "--hysteresis",
                    action='store', default=0.15, type=float,
                    help="Derefinement safety coefficient.")
parser.add_argument('-r', '--ref-levels',
                    action='store', default=0, type=int,
                    help="Number of inital uniform refinement")
parser.add_argument("-l", "--nc-limit",
                    action='store', default=3, type=int,
                    help="Maximum level of hanging nodes.")
parser.add_argument('-tf', '--t-final',
                    action='store', default=1.0, type=float,
                    help="Final time; start time is 0.")
parser.add_argument('-vis', '--visualization',
                    action='store_true', default=True,
                    help='Enable GLVis visualization')

args = parser.parse_args()
ref_levels = args.ref_levels
problem = args.problem
nfeatures = args.nfeatures
order = args.order
max_elem_error = args.max_err
hysteresis = args.hysteresis
nc_limit = args.nc_limit
t_final = args.t_final
visualization = args.visualization
parser.print_options(args)


alpha = 0.02


def front(x, y, z, t, dim):
    r = sqrt(x**2 + y**2 + z**2)
    return exp(-0.5 * ((r - t)/alpha) ** 2)


def ball(x, y, z, t, dim):
    r = sqrt(x**2 + y**2 + z**2)
    return -arctan(2.*(r - t)/alpha)


def front_laplace(x, y, z, t, dim):
    x2 = x**2
    y2 = y**2
    z2 = z**2
    t2 = t**2
    a2 = alpha**2
    a4 = a2**2
    r = sqrt(x2 + y2 + z2)
    ret = (-exp(- 0.5 * ((r - t)/alpha)**2) / a4 *
           (-2.*t*(x2 + y2 + z2 - (dim-1)*a2/2.)/r + x2 + y2 + z2 + t2 - dim*a2))
    return ret


def ball_laplace(x, y, z, t, dim):
    x2 = x**2
    y2 = y**2
    z2 = z**2
    t2 = t**2
    a2 = alpha**2
    a4 = a2**2
    r = sqrt(x2 + y2 + z2)
    den = (-a2 - 4.*(x2 + y2 + z2 - 2*r*t) - t2)**2

    if dim == 2:
        return 2*alpha*(a2 + t2 - 4*x2 - 4*y2)/r/den
    return 4*alpha*(a2 + t2 - 4.*r*t)/r/den


def composite_func(pt, t, f0, f1):
    dim = len(pt)
    x = pt[0]
    y = pt[1]
    z = 0.0
    if dim == 3:
        z = pt[2]
    if (problem == 0):
        if (nfeatures <= 1):
            return f0(x, y, z, t, dim)
        else:
            i = np.arange(nfeatures)
            x0 = 0.5*cos(2 * pi * i / nfeatures)
            y0 = 0.5*sin(2 * pi * i / nfeatures)
            return np.sum(f0(x - x0, y - y0, z, t, dim))
    else:
        i = np.arange(nfeatures)
        x0 = 0.5*cos(2 * pi * i / nfeatures + pi*t)
        y0 = 0.5*sin(2 * pi * i / nfeatures + pi*t)
        return np.sum(f1(x - x0, y - y0, z, 0.25, dim))


if HAS_NUMBA:
    jit_module(nopython=True, error_model="numpy")

    @mfem.jit.scalar(td=True)
    def bdr(pt, t):
        return composite_func(pt, t, front, ball)

    @mfem.jit.scalar(td=True)
    def rhs(pt, t):
        return composite_func(pt, t, front_laplace, ball_laplace)
else:
    class BdrCoefficient(mfem.PyCoefficientT):
        def EvalValue(self, pt, t):
            return composite_func(pt, t, front, ball)

    class RhsCoefficient(mfem.PyCoefficientT):
        def EvalValue(self, pt, t):
            # print 'rhs', composite_func(pt, t, front_laplace, ball_laplace)
            return composite_func(pt, t, front_laplace, ball_laplace)
    bdr = BdrCoefficient()
    rhs = RhsCoefficient()


def UpdateProblem(mesh, fespace, x, a, b):
    # Update the space: recalculate the number of DOFs and construct a matrix
    # that will adjust any GridFunctions to the new mesh state.
    fespace.Update()

    # Interpolate the solution on the new mesh by applying the transformation
    # matrix computed in the finite element space. Multiple GridFunctions could
    # be updated here.
    x.Update()

    # Free any transformation matrices to save memory.
    fespace.UpdatesFinished()

    # Inform the linear and bilinear forms that the space has changed.
    a.Update()
    b.Update()


# 2. Read the mesh from the given mesh file on all processors. We can handle
#    triangular, quadrilateral, tetrahedral, hexahedral, surface and volume
#    meshes with the same code
meshfile = expanduser(join(dirname(__file__), '..', 'data', args.mesh))
mesh = mfem.Mesh(meshfile, 1, 1)
dim = mesh.Dimension()
sdim = mesh.SpaceDimension()

# 3. Project a NURBS mesh to a piecewise-quadratic curved mesh. Make sure
#    that the mesh is non-conforming if it has quads or hexes and refine it
if (mesh.NURBSext):
    mesh.UniformRefinement()
    if ref_levels > 0:
        ref_levels = ref_levels-1
    mesh.SetCurvature(2)

mesh.EnsureNCMesh()
for l in range(ref_levels):
    mesh.UniformRefinement()

# 4. All boundary attributes will be used for essential (Dirichlet) BC
ess_bdr = intArray(mesh.bdr_attributes.Max())
ess_bdr.Assign(1)

# 5. Define a finite element space on the mesh. The polynomial order is one
#   (linear) by default, but this can be changed on the command line.
fec = mfem.H1_FECollection(order, dim)
fespace = mfem.FiniteElementSpace(mesh, fec)

# 6. As in Example 1p, we set up bilinear and linear forms corresponding to
#    the Laplace problem -\Delta u = 1. We don't assemble the discrete
#    problem yet, this will be done in the inner loop.
a = mfem.BilinearForm(fespace)
b = mfem.LinearForm(fespace)

one = mfem.ConstantCoefficient(1.0)

integ = mfem.DiffusionIntegrator(one)
a.AddDomainIntegrator(integ)
b.AddDomainIntegrator(mfem.DomainLFIntegrator(rhs))

# 7. The solution vector x and the associated finite element grid function
#    will be maintained over the AMR iterations.
x = mfem.GridFunction(fespace)

# 8. Connect to GLVis.
if visualization:
    sout = mfem.socketstream("localhost", 19916)
    sout.precision(8)

# 9. As in Example 6, we set up a Zienkiewicz-Zhu estimator that will be
#    used to obtain element error indicators. The integrator needs to
#    provide the method ComputeElementFlux. The smoothed flux space is a
#    vector valued H1 space here.
flux_fespace = mfem.FiniteElementSpace(mesh, fec, sdim)

# own_flux_fes = False indicate flux_fespace is passed by reference
# this is actually default action, but for the sake of explanaiton
# it is explicitly set. If you want to pass pointer use own_flux_fes = True
estimator = mfem.ZienkiewiczZhuEstimator(integ, x, flux_fespace,
                                         own_flux_fes=False)

# 10. As in Example 6, we also need a refiner. This time the refinement
#     strategy is based on a fixed threshold that is applied locally to each
#     element. The global threshold is turned off by setting the total error
#     fraction to zero. We also enforce a maximum refinement ratio between
#     adjacent elements.
refiner = mfem.ThresholdRefiner(estimator)
refiner.SetTotalErrorFraction(0.0)
refiner.SetLocalErrorGoal(max_elem_error)
refiner.PreferConformingRefinement()
refiner.SetNCLimit(nc_limit)

# 11. A derefiner selects groups of elements that can be coarsened to form
#     a larger element. A conservative enough threshold needs to be set to
#     prevent derefining elements that would immediately be refined again.
derefiner = mfem.ThresholdDerefiner(estimator)
derefiner.SetThreshold(hysteresis * max_elem_error)
derefiner.SetNCLimit(nc_limit)

# 12. The outer time loop. In each iteration we update the right hand side,
#     solve the problem on the current mesh, visualize the solution and
#     refine the mesh as many times as necessary. Then we derefine any
#     elements which have very small errors.
x.Assign(0.0)
time = 0.0
while (time < t_final + 1e-10):
    print("Time " + str(time) + "\n\nRefinement:")
    bdr.SetTime(time)
    rhs.SetTime(time)

    # Make sure errors will be recomputed in the following.
    refiner.Reset()
    derefiner.Reset()

    # 13. The inner refinement loop. At the end we want to have the current
    #     time step resolved to the prescribed tolerance in each element.
    ref_it = 0
    while(True):
        ref_it = ref_it + 1
        print("Iteration: " + str(ref_it) + ", number of unknowns: "
              + str(fespace.GetVSize()))

        # 14. Recompute the field on the current mesh: assemble the stiffness
        #     matrix and the right-hand side.
        a.Assemble()
        b.Assemble()

        # 15. Project the exact solution to the essential boundary DOFs.
        x.ProjectBdrCoefficient(bdr, ess_bdr)
        # 16. Create and solve the linear system.
        ess_tdof_list = intArray()
        fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

        A = mfem.OperatorPtr()
        B = mfem.Vector()
        X = mfem.Vector()
        a.FormLinearSystem(ess_tdof_list, x, b, A, X, B)

        AA = mfem.OperatorHandle2SparseMatrix(A)
        M = mfem.GSSmoother(AA)
        mfem.PCG(AA, M, B, X, 0, 500, 1e-12, 0.0)

        # 17. Extract the local solution on each processor.
        a.RecoverFEMSolution(X, b, x)

        # 18. Send the solution by socket to a GLVis server
        if visualization:
            sout.precision(8)
            sout.send_solution(mesh, x)

        # 19. Apply the refiner on the mesh. The refiner calls the error
        #     estimator to obtain element errors, then it selects elements to
        #     be refined and finally it modifies the mesh. The Stop() method
        #     determines if all elements satisfy the local threshold.
        refiner.Apply(mesh)
        if refiner.Stop():
            break

        # 20. Update the space and interpolate the solution.
        UpdateProblem(mesh, fespace, x, a, b)
    # 21. Use error estimates from the last inner iteration to check for
    #     possible derefinements. The derefiner works similarly as the
    #     refiner. The errors are not recomputed because the mesh did not
    #     change (and also the estimator was not Reset() at this time).
    if derefiner.Apply(mesh):
        print("Derefined elements.")
        #  22. Update the space and interpolate the solution.
        UpdateProblem(mesh, fespace, x, a, b)
    a.Update()
    b.Update()
    time = time + 0.01



'''
   MFEM example 10

      This examples solves a time dependent nonlinear elasticity
      problem of the form dv/dt = H(x) + S v, dx/dt = v, where H is a
      hyperelastic model and S is a viscosity operator of Laplacian
      type.

      refinement loop. 
      See c++ version in the MFEM library for more detail 
'''
import sys
from mfem.common.arg_parser import ArgParser
import mfem.ser as mfem
from mfem.ser import intArray, add_vector, Add
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sqrt, pi, cos, sin, hypot, arctan2
from scipy.special import erfc

parser = ArgParser(description='Ex10')
parser.add_argument('-m', '--mesh',
                    default='beam-quad.mesh',
                    action='store', type=str,
                    help='Mesh file to use.')
parser.add_argument('-r', '--refine-serial',
                    action='store', default=2, type=int,
                    help="Number of times to refine the mesh uniformly before parallel")
parser.add_argument('-o', '--order',
                    action='store', default=2, type=int,
                    help="Finite element order (polynomial degree)")
help_ode = "\n".join(["ODE solver: 1 - Backward Euler, 2 - SDIRK2, 3 - SDIRK3",
                      "\t11 - Forward Euler, 12 - RK2",
                      "\t13 - RK3 SSP, 14 - RK4."])
parser.add_argument('-s', '--ode-solver',
                    action='store', default=3, type=int,
                    help=help_ode)
parser.add_argument('-tf', '--t-final',
                    action='store', default=300.0, type=float,
                    help="Final time; start time is 0.")
parser.add_argument('-dt', '--time-step',
                    action='store', default=3.0, type=float,
                    help="Time step")
parser.add_argument("-v", "--viscosity",
                    action='store', default=1e-2, type=float,
                    help="Viscosity coefficient.")
parser.add_argument("-mu", "--shear-modulus",
                    action='store', default=0.25, type=float,
                    help="Shear modulus in the Neo-Hookean hyperelastic model.")
parser.add_argument("-K", "--bulk-modulus",
                    action='store', default=5.0, type=float,
                    help="Bulk modulus in the Neo-Hookean hyperelastic model.")
parser.add_argument('-vis', '--visualization',
                    action='store_true', default=True,
                    help='Enable GLVis visualization')
parser.add_argument("-vs", "--visualization-steps",
                    action='store', default=1,  type=int,
                    help="Visualize every n-th timestep.")
args = parser.parse_args()

ref_levels = args.refine_serial
order = args.order
ode_solver_type = args.ode_solver
t_final = args.t_final
dt = args.time_step
visc = args.viscosity
mu = args.shear_modulus
K = args.bulk_modulus
visualization = args.visualization
vis_steps = args.visualization_steps
meshfile = expanduser(
    join(dirname(__file__), '..', 'data', args.mesh))

parser.print_options(args)

mesh = mfem.Mesh(meshfile, 1, 1)
dim = mesh.Dimension()
#        self.solver.SetOperator(M)
if ode_solver_type == 1:
    ode_solver = BackwardEulerSolver()
elif ode_solver_type == 2:
    ode_solver = mfem.SDIRK23Solver(2)
elif ode_solver_type == 3:
    ode_solver = mfem.SDIRK33Solver()
elif ode_solver_type == 11:
    ode_solver = ForwardEulerSolver()
elif ode_solver_type == 12:
    ode_solver = mfem.RK2Solver(0.5)
elif ode_solver_type == 13:
    ode_solver = mfem.RK3SSPSolver()
elif ode_solver_type == 14:
    ode_solver = mfem.RK4Solver()
elif ode_solver_type == 22:
    ode_solver = mfem.ImplicitMidpointSolver()
elif ode_solver_type == 23:
    ode_solver = mfem.SDIRK23Solver()
elif ode_solver_type == 24:
    ode_solver = mfem.SDIRK34Solver()
else:
    print("Unknown ODE solver type: " + str(ode_solver_type))
    exit

for lev in range(ref_levels):
    mesh.UniformRefinement()

# 5. Define the vector finite element spaces representing the mesh
#    deformation x, the velocity v, and the initial configuration, x_ref.
#    Define also the elastic energy density, w, which is in a discontinuous
#    higher-order space. Since x and v are integrated in time as a system,
#    we group them together in block vector vx, with offsets given by the
#    fe_offset array.
fec = mfem.H1_FECollection(order, dim)
fespace = mfem.FiniteElementSpace(mesh, fec, dim)

fe_size = fespace.GetVSize()
print("Number of velocity/deformation unknowns: " + str(fe_size))
fe_offset = intArray([0, fe_size, 2*fe_size])

vx = mfem.BlockVector(fe_offset)
x = mfem.GridFunction()
v = mfem.GridFunction()
v.MakeRef(fespace, vx.GetBlock(0), 0)
x.MakeRef(fespace, vx.GetBlock(1), 0)

x_ref = mfem.GridFunction(fespace)
mesh.GetNodes(x_ref)

w_fec = mfem.L2_FECollection(order + 1, dim)
w_fespace = mfem.FiniteElementSpace(mesh, w_fec)
w = mfem.GridFunction(w_fespace)

# 6. Set the initial conditions for v and x, and the boundary conditions on
#    a beam-like mesh (see description above).


class InitialVelocity(mfem.VectorPyCoefficient):
    def EvalValue(self, x):
        dim = len(x)
        s = 0.1/64.

        v = np.zeros(len(x))
        v[-1] = s*x[0]**2*(8.0-x[0])
        v[0] = -s*x[0]**2
        return v


class InitialDeformation(mfem.VectorPyCoefficient):
    def EvalValue(self, x):
        return x.copy()


velo = InitialVelocity(dim)
v.ProjectCoefficient(velo)
deform = InitialDeformation(dim)
x.ProjectCoefficient(deform)

ess_bdr = intArray(fespace.GetMesh().bdr_attributes.Max())
ess_bdr.Assign(0)
ess_bdr[0] = 1

# 7. Define HyperelasticOperator and initialize it
#    the initial energies.


class ElasticEnergyCoefficient(mfem.PyCoefficient):
    def __init__(self, model, x):
        self.x = x
        self.model = model
        self.J = mfem.DenseMatrix()
        mfem.PyCoefficient.__init__(self)

    def Eval(self, T, ip):
        self.model.SetTransformation(T)
        self.x.GetVectorGradient(T, self.J)
        # T.Jacobian().Print()
        # print self.x.GetDataArray()
        # self.J.Print()
        return self.model.EvalW(self.J)/(self.J.Det())


class ReducedSystemOperator(mfem.PyOperator):
    def __init__(self, M, S, H):
        mfem.PyOperator.__init__(self, M.Height())
        self.M = M
        self.S = S
        self.H = H
        self.Jacobian = None
        h = M.Height()
        self.w = mfem.Vector(h)
        self.z = mfem.Vector(h)
        self.dt = 0.0
        self.v = None
        self.x = None

    def SetParameters(self, dt, v, x):
        self.dt = dt
        self.v = v
        self.x = x

    def Mult(self, k, y):
        add_vector(self.v, self.dt, k, self.w)
        add_vector(self.x, self.dt, self.w, self.z)
        self.H.Mult(self.z, y)
        self.M.AddMult(k, y)
        self.S.AddMult(self.w, y)

    def GetGradient(self, k):
        Jacobian = Add(1.0, self.M.SpMat(), self.dt, self.S.SpMat())
        self.Jacobian = Jacobian
        add_vector(self.v, self.dt, k, self.w)
        add_vector(self.x, self.dt, self.w, self.z)
        grad_H = self.H.GetGradientMatrix(self.z)

        Jacobian.Add(self.dt**2, grad_H)
        return Jacobian


class HyperelasticOperator(mfem.PyTimeDependentOperator):
    def __init__(self, fespace, ess_bdr, visc, mu, K):
        mfem.PyTimeDependentOperator.__init__(self, 2*fespace.GetVSize(), 0.0)

        rel_tol = 1e-8
        skip_zero_entries = 0
        ref_density = 1.0
        self.z = mfem.Vector(self.Height()//2)
        self.fespace = fespace
        self.viscosity = visc

        M = mfem.BilinearForm(fespace)
        S = mfem.BilinearForm(fespace)
        H = mfem.NonlinearForm(fespace)
        self.M = M
        self.H = H
        self.S = S

        rho = mfem.ConstantCoefficient(ref_density)
        M.AddDomainIntegrator(mfem.VectorMassIntegrator(rho))
        M.Assemble(skip_zero_entries)
        M.EliminateEssentialBC(ess_bdr)
        M.Finalize(skip_zero_entries)

        M_solver = mfem.CGSolver()
        M_prec = mfem.DSmoother()
        M_solver.iterative_mode = False
        M_solver.SetRelTol(rel_tol)
        M_solver.SetAbsTol(0.0)
        M_solver.SetMaxIter(30)
        M_solver.SetPrintLevel(0)
        M_solver.SetPreconditioner(M_prec)
        M_solver.SetOperator(M.SpMat())

        self.M_solver = M_solver
        self.M_prec = M_prec

        model = mfem.NeoHookeanModel(mu, K)
        H.AddDomainIntegrator(mfem.HyperelasticNLFIntegrator(model))
        H.SetEssentialBC(ess_bdr)
        self.model = model

        visc_coeff = mfem.ConstantCoefficient(visc)
        S.AddDomainIntegrator(mfem.VectorDiffusionIntegrator(visc_coeff))
        S.Assemble(skip_zero_entries)
        S.EliminateEssentialBC(ess_bdr)
        S.Finalize(skip_zero_entries)

        self.reduced_oper = ReducedSystemOperator(M, S, H)

        J_prec = mfem.DSmoother(1)
        J_minres = mfem.MINRESSolver()
        J_minres.SetRelTol(rel_tol)
        J_minres.SetAbsTol(0.0)
        J_minres.SetMaxIter(300)
        J_minres.SetPrintLevel(-1)
        J_minres.SetPreconditioner(J_prec)

        self.J_solver = J_minres
        self.J_prec = J_prec

        newton_solver = mfem.NewtonSolver()
        newton_solver.iterative_mode = False
        newton_solver.SetSolver(self.J_solver)
        newton_solver.SetOperator(self.reduced_oper)
        newton_solver.SetPrintLevel(1)  # print Newton iterations
        newton_solver.SetRelTol(rel_tol)
        newton_solver.SetAbsTol(0.0)
        newton_solver.SetMaxIter(10)
        self.newton_solver = newton_solver

    def Mult(self, vx, vx_dt):
        sc = self.Height()//2
        v = mfem.Vector(vx, 0,  sc)
        x = mfem.Vector(vx, sc,  sc)
        dv_dt = mfem.Vector(dvx_dt, 0, sc)
        dx_dt = mfem.Vector(dvx_dt, sc,  sc)
        self.H.Mult(x, z)
        if (self.viscosity != 0.0):
            S.AddMult(v, z)
        z.Neg()
        M_solver.Mult(z, dv_dt)
        dx_dt = v
#        Print(vx.Size())

    def ImplicitSolve(self, dt, vx, dvx_dt):
        sc = self.Height()//2
        v = mfem.Vector(vx, 0,  sc)
        x = mfem.Vector(vx, sc,  sc)
        dv_dt = mfem.Vector(dvx_dt, 0, sc)
        dx_dt = mfem.Vector(dvx_dt, sc,  sc)

        # By eliminating kx from the coupled system:
        # kv = -M^{-1}*[H(x + dt*kx) + S*(v + dt*kv)]
        # kx = v + dt*kv
        # we reduce it to a nonlinear equation for kv, represented by the
        # backward_euler_oper. This equation is solved with the newton_solver
        # object (using J_solver and J_prec internally).
        self.reduced_oper.SetParameters(dt, v, x)
        zero = mfem.Vector()  # empty vector is interpreted as
        # zero r.h.s. by NewtonSolver
        self.newton_solver.Mult(zero, dv_dt)
        add_vector(v, dt, dv_dt, dx_dt)

    def ElasticEnergy(self, x):
        return self.H.GetEnergy(x)

    def KineticEnergy(self, v):
        return 0.5*self.M.InnerProduct(v, v)

    def GetElasticEnergyDensity(self, x, w):
        w_coeff = ElasticEnergyCoefficient(self.model, x)
        w.ProjectCoefficient(w_coeff)


oper = HyperelasticOperator(fespace, ess_bdr, visc, mu, K)
ee0 = oper.ElasticEnergy(x)
ke0 = oper.KineticEnergy(v)

print("initial elastic energy (EE) = " + str(ee0))
print("initial kinetic energy (KE) = " + str(ke0))
print("initial   total energy (TE) = " + str(ee0 + ke0))

# 8. Perform time-integration (looping over the time iterations, ti, with a
#    time-step dt).
ode_solver.Init(oper)
t = 0.
ti = 1
last_step = False
while not last_step:
    if (t + dt >= t_final - dt/2):
        last_step = True

    t, dt = ode_solver.Step(vx, t, dt)

    if (last_step or (ti % vis_steps) == 0):
        ee = oper.ElasticEnergy(x)
        ke = oper.KineticEnergy(v)

        text = ("step " + str(ti) + ", t = " + str(t) + ", EE = " +
                str(ee) + ", KE = " + str(ke) +
                ", dTE = " + str((ee+ke)-(ee0+ke0)))

        print(text)
    ti = ti + 1
#
# if i translate c++ line-by-line, ti seems the second swap does not work...
#
nodes = x
owns_nodes = 0
nodes, owns_nodes = mesh.SwapNodes(nodes, owns_nodes)
mesh.Print('deformed.mesh', 8)
mesh.SwapNodes(nodes, owns_nodes)
v.Save('velocity.sol', 8)
oper.GetElasticEnergyDensity(x, w)
w.Save('elastic_energy.sol',  8)



from __future__ import print_function
import os
import sys

def run_test():
    print("Test point module")
    
    pt = mfem.Point(3)
    pt.SetAttribute(1)
    #print(pt.GetVertices(), pt.GetAttribute())
    assert  pt.GetVertices() == 3, "failed to set attribute (must be 3)"
    assert  pt.GetAttribute() == 1, "failed to set attribute (must be 1)"

if __name__=='__main__':
    if len(sys.argv) > 1 and sys.argv[1] == '-p':   
        import mfem.par as mfem
    else:
        import mfem.ser as mfem
    
    run_test()





from __future__ import print_function

def nicePrint(*s):
    from mpi4py import MPI
    
    comm     = MPI.COMM_WORLD     
    nproc = MPI.COMM_WORLD.size
    myid     = MPI.COMM_WORLD.rank

    comm.Barrier() 
    for i in range(nproc):
        comm.Barrier()        
        if i == myid: print("[ID:"+str(myid)+"]" + ' '.join([str(ss) for ss in s]))
        comm.Barrier()
    
def niceCall(m, *args, **kargs):        
    from mpi4py import MPI
    
    comm     = MPI.COMM_WORLD     
    nproc = MPI.COMM_WORLD.size
    myid     = MPI.COMM_WORLD.rank

    comm.Barrier() 
    for i in range(nproc):
        comm.Barrier()        
        if i == myid:
            print("[ID:"+ str(myid) +"]")
            m(*args, **kargs)
        comm.Barrier()
        



'''
   MFEM example 40p (converted from ex40p.cpp)

   See c++ version in the MFEM library for more detail

   Sample runs: mpirun -np  4 python ex40.py -step 10.0 -gr 2.0
                mpirun -np  4 python ex40.py -step 10.0 -gr 2.0 -o 3 -r 1
                mpirun -np  4 python ex40.py -step 10.0 -gr 2.0 -r 4 -m ../data/l-shape.mesh
                mpirun -np  4 python ex40.py -step 10.0 -gr 2.0 -r 2 -m ../data/fichera.mesh

   Description: This example code demonstrates how to use MFEM to solve the
                eikonal equation,

                        |∇𝑢| = 1 in Ω,  𝑢 = 0 on ∂Ω.

                The viscosity solution of this problem coincides with the unique optimum
                of the nonlinear program

                     maximize ∫_Ω 𝑢 d𝑥 subject to |∇𝑢| ≤ 1 in Ω, 𝑢 = 0 on ∂Ω,    (⋆)

                which is the foundation for method implemented below.

                Following the proximal Galerkin methodology [1,2] (see also Example
                36), we construct a Legendre function for the closed unit ball
                𝐵₁ := {𝑥 ∈ Rⁿ | |𝑥| ≤ 1}. Our choice is the Hellinger entropy,

                      R(𝑥) = −( 1 − |𝑥|² )^{1/2},

                although other choices are possible, each leading to a slightly
                different algorithm. We then adaptively regularize the optimization
                problem (⋆) with the Bregman divergence of the Hellinger entropy,

                   maximize  ∫_Ω 𝑢 d𝑥 - αₖ⁻¹ D(∇𝑢,∇𝑢ₖ₋₁)  subject to  𝑢 = 0 on Ω.

                This results in a sequence of functions ( 𝜓ₖ , 𝑢ₖ ),

                        𝑢ₖ → 𝑢,    𝜓ₖ/|𝜓ₖ| → ∇𝑢    as k → ∞,

                defined by the nonlinear saddle-point problems

                 Find 𝜓ₖ ∈ H(div,Ω) and 𝑢ₖ ∈ L²(Ω) such that
                 ( (∇R)⁻¹(𝜓ₖ) , τ ) + ( 𝑢ₖ , ∇⋅τ ) = 0                     ∀ τ ∈ H(div,Ω)
                 ( ∇⋅𝜓ₖ , v )                     = ( ∇⋅𝜓ₖ₋₁ - αₖ , v )    ∀ v ∈ L²(Ω)

                where (∇R)⁻¹(𝜓) = 𝜓 / ( 1 + |𝜓|² )^{1/2} and αₖ = α₀rᵏ, where r ≥ 1
                is a prescribed growth rate. (r = 1 is the most stable.) The
                saddle-point problems are solved using a damped quasi-Newton method
                with a tunable regularization parameter 0 ≤ ϵ << 1.

                [1] Keith, B. and Surowiec, T. (2024) Proximal Galerkin: A structure-
                    preserving finite element method for pointwise bound constraints.
                    Foundations of Computational Mathematics, 1–97.
                [2] Dokken, J., Farrell, P., Keith, B., Papadopoulos, I., and
                    Surowiec, T. (2025) The latent variable proximal point algorithm
                    for variational problems with inequality constraints. (To appear.)

'''
import os
from os.path import expanduser, join
import numpy as np

import mfem.par as mfem

from mpi4py import MPI
num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
smyid = '.'+'{:0>6d}'.format(myid)


def run(meshfile="",
        order=1,
        max_it=5,
        ref_levels=3,
        alpha=1.0,
        growth_rate=1.0,
        newton_scaling=0.8,
        eps=1e-6,
        tol=1e-4,
        max_alpha=1e2,
        max_psi=1e2,
        eps2=1e-1,
        visualization=True):

    # 2. Read the mesh from the mesh file.
    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()
    sdim = mesh.SpaceDimension()

    # 3. Postprocess the mesh.
    # 3A. Refine the mesh to increase the resolution.
    for i in range(ref_levels):
        mesh.UniformRefinement()

    # 3B. Interpolate the geometry after refinement to control geometry error.
    # NOTE: Minimum second-order interpolation is used to improve the accuracy.
    curvature_order = max(order, 2)
    mesh.SetCurvature(curvature_order)

    # 3C. Compute the maximum mesh size.
    hmax = 0.0
    for i in range(mesh.GetNE()):
        hmax = max(mesh.GetElementSize(i, 1), hmax)

    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
    del mesh

    # 4. Define the necessary finite element spaces on the mesh.
    RTfec = mfem.RT_FECollection(order, dim)
    RTfes = mfem.ParFiniteElementSpace(pmesh, RTfec)

    L2fec = mfem.L2_FECollection(order, dim)
    L2fes = mfem.ParFiniteElementSpace(pmesh, L2fec)

    dofs_rt = RTfes.GlobalTrueVSize()
    dofs_l2 = L2fes.GlobalTrueVSize()
    if myid == 0:
        print("Number of H(div) dofs: " + str(dofs_rt))
        print("Number of L² dofs: " + str(dofs_l2))

    # 5. Define the offsets for the block matrices
    offsets = mfem.intArray([0, RTfes.GetVSize(), L2fes.GetVSize()])
    offsets.PartialSum()
    toffsets = mfem.intArray([0, RTfes.GetTrueVSize(), L2fes.GetTrueVSize()])
    toffsets.PartialSum()

    x = mfem.BlockVector(offsets)
    rhs = mfem.BlockVector(offsets)
    x.Assign(0.0)
    rhs.Assign(0.0)
    tx = mfem.BlockVector(toffsets)
    trhs = mfem.BlockVector(toffsets)
    tx.Assign(0.0)
    trhs.Assign(0.0)

    # 6. Define the solution vectors as a finite element grid functions
    #    corresponding to the fespaces.

    u_gf = mfem.ParGridFunction()
    delta_psi_gf = mfem.ParGridFunction()
    delta_psi_gf.MakeRef(RTfes, x, offsets[0])
    u_gf.MakeRef(L2fes, x, offsets[1])

    psi_old_gf = mfem.ParGridFunction(RTfes)
    psi_gf = mfem.ParGridFunction(RTfes)
    u_old_gf = mfem.ParGridFunction(L2fes)

    # 7. Define initial guesses for the solution variables.
    delta_psi_gf.Assign(0.0)
    psi_gf.Assign(0.0)
    u_gf.Assign(0.0)
    psi_old_gf.Assign(0.0)
    u_old_gf.Assign(0.0)

    # 8. Prepare for glvis output.
    # 14. Send the solution by socket to a GLVis server.
    if visualization:
        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock.precision(8)

    # 9. Coefficients to be used later.
    zero_cf = mfem.ConstantCoefficient(0.0)
    one_cf = mfem.ConstantCoefficient(1.0)
    zero_vec = mfem.Vector([0.0]*sdim)
    zero_vec_cf = mfem.VectorConstantCoefficient(zero_vec)
    neg_alpha_cf = mfem.ConstantCoefficient(-1.0*alpha)

    psigf_cf = mfem.VectorGridFunctionCoefficient(psi_gf)

    @mfem.jit.vector(shape=(sdim,), dependency=(psigf_cf,))
    def Z(_ptx, psi_vals):
        norm = np.linalg.norm(psi_vals)
        phi = 1.0 / np.sqrt(1.0 + norm*norm)
        vvec = psi_vals*phi
        return vvec

    @mfem.jit.matrix(shape=(sdim, sdim), dependency=(psigf_cf,))
    def DZ(_ptx, psi_vals):
        norm = np.linalg.norm(psi_vals)
        phi = 1.0 / np.sqrt(1.0 + norm*norm)

        kmat = np.zeros((sdim, sdim))
        for i in range(sdim):
            kmat[i, i] = phi + eps
            for j in range(sdim):
                kmat[i, j] -= psi_vals[i]*psi_vals[j] * phi**3
        return kmat

    neg_Z = mfem.ScalarVectorProductCoefficient(-1.0, Z)
    div_psi_cf = mfem.DivergenceGridFunctionCoefficient(psi_gf)
    div_psi_old_cf = mfem.DivergenceGridFunctionCoefficient(psi_old_gf)
    psi_old_minus_psi = mfem.SumCoefficient(
        div_psi_old_cf, div_psi_cf, 1.0, -1.0)

    # 10. Assemble constant matrices/vectors to avoid reassembly in the loop.
    b0 = mfem.ParLinearForm()
    b1 = mfem.ParLinearForm()
    b0.MakeRef(RTfes, rhs.GetBlock(0), 0)
    b1.MakeRef(L2fes, rhs.GetBlock(1), 0)

    b0.AddDomainIntegrator(mfem.VectorFEDomainLFIntegrator(neg_Z))
    b1.AddDomainIntegrator(mfem.DomainLFIntegrator(neg_alpha_cf))
    b1.AddDomainIntegrator(mfem.DomainLFIntegrator(psi_old_minus_psi))

    a00 = mfem.ParBilinearForm(RTfes)
    a00.AddDomainIntegrator(mfem.VectorFEMassIntegrator(DZ))

    a10 = mfem.ParMixedBilinearForm(RTfes, L2fes)
    a10.AddDomainIntegrator(mfem.VectorFEDivergenceIntegrator())
    a10.Assemble()
    a10.Finalize()
    A10 = a10.ParallelAssemble()
    A01 = A10.Transpose()

    vol_form = mfem.ParLinearForm(L2fes)
    one_gf = mfem.ParGridFunction(L2fes)
    one_gf.Assign(1.0)
    vol_form.AddDomainIntegrator(mfem.DomainLFIntegrator(one_cf))
    vol_form.Assemble()
    domain_volume = vol_form(one_gf)

    # 11. Iterate.
    total_iterations = 0
    increment_u = 0.1
    u_tmp = mfem.ParGridFunction(L2fes)

    for k in range(max_it):
        u_tmp.Assign(u_old_gf)

        if myid == 0:
            print("\nOUTER ITERATION " + str(k+1))

        for j in range(5):
            total_iterations += 1

            b0.Assemble()
            b0.ParallelAssemble(trhs.GetBlock(0))

            b1.Assemble()
            b1.ParallelAssemble(trhs.GetBlock(1))

            a00.Assemble(0)
            a00.Finalize(0)
            A00 = a00.ParallelAssemble()

            # Construct Schur-complement preconditioner
            A00_diag = mfem.HypreParVector(MPI.COMM_WORLD,
                                           A00.GetGlobalNumRows(),
                                           A00.GetRowStarts())
            A00.GetDiag(A00_diag)
            S_tmp = mfem.HypreParMatrix(A01)
            S_tmp.InvScaleRows(A00_diag)
            A00_diag.Reciprocal()
            S = mfem.ParMult(A10, S_tmp, True)

            # Python note:
            #    owns_blocks should be 0 in Python
            #    because wrapper class will delete blocks
            prec = mfem.BlockDiagonalPreconditioner(toffsets)
            prec.owns_blocks = 0

            P00 = mfem.HypreBoomerAMG(A00)
            P00.SetPrintLevel(0)
            P11 = mfem.HypreBoomerAMG(S)
            P11.SetPrintLevel(0)
            prec.SetDiagonalBlock(0, P00)
            prec.SetDiagonalBlock(1, P11)

            A = mfem.BlockOperator(toffsets)
            A.SetBlock(0, 0, A00)
            A.SetBlock(1, 0, A10)
            A.SetBlock(0, 1, A01)

            minres = mfem.MINRESSolver(MPI.COMM_WORLD)
            minres.SetPrintLevel(-1)
            minres.SetRelTol(1e-12)
            minres.SetMaxIter(10000)
            minres.SetOperator(A)
            minres.SetPreconditioner(prec)
            minres.Mult(trhs, tx)

            del S
            del A00

            delta_psi_gf.SetFromTrueDofs(tx.GetBlock(0))
            u_gf.SetFromTrueDofs(tx.GetBlock(1))

            u_tmp -= u_gf  # u_tmp = u_tmp - u_gf

            Newton_update_size = u_tmp.ComputeL2Error(zero_cf)
            u_tmp.Assign(u_gf)

            # Damped Newton update
            psi_gf.Add(newton_scaling, delta_psi_gf)
            a00.Update()

            if visualization:
                sol_sock << "parallel " << num_procs << " " << myid << "\n"
                sol_sock << "solution\n" << pmesh << u_gf
                sol_sock << "window_title 'Discrete solution'"
                sol_sock.flush()

            if myid == 0:
                print("Newton_update_size = " +
                      "{:g}".format(Newton_update_size))

            if newton_scaling*Newton_update_size < increment_u:
                break

        u_tmp.Assign(u_gf)
        u_tmp -= u_old_gf
        increment_u = u_tmp.ComputeL2Error(zero_cf)

        if myid == 0:
            print("Number of Newton iterations = " + str(j+1))
            print("Increment (|| uₕ - uₕ_prvs||) = " +
                  "{:g}".format(increment_u))

        u_old_gf.Assign(u_gf)
        psi_old_gf.Assign(psi_gf)
        alpha *= max(growth_rate, 1.0)

        #  Safeguard 1: Stop alpha from growing too large
        alpha = min(alpha, max_alpha)

        # Safeguard 2: Stop |ψ| from growing too large
        norm_psi = psi_old_gf.ComputeL1Error(zero_vec_cf)/domain_volume
        if norm_psi > max_psi:
            # Additional entropy regularization
            neg_alpha_cf.constant = -alpha/(1.0 + eps2 * alpha * hmax)
            psi_old_minus_psi.SetAlpha(1.0/(1.0 + eps2 * alpha * hmax))
        else:
            neg_alpha_cf.constant = -alpha

        if increment_u < tol or k == max_it-1:
            break

        alpha *= max(growth_rate, 1.0)
        neg_alpha_cf.constant = -alpha

    if myid == 0:
        print("\n Outer iterations: " + str(k+1) +
              "\n Total iterations: " + str(total_iterations) +
              "\n Total dofs:       " + str(RTfes.GetTrueVSize() + L2fes.GetTrueVSize()))


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex40 (Eikonal queation)')
    parser.add_argument('-m', '--mesh',
                        default='star.mesh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree).")
    parser.add_argument('-r', '--refs',
                        action='store', default=3, type=int,
                        help="Number of h-refinements.")
    parser.add_argument('-mi', '--max-it',
                        action='store', default=5, type=int,
                        help="Maximum number of iterations")
    parser.add_argument("-tol", "--tol",
                        action='store', default=1e-4, type=float,
                        help="Stopping criteria based on the difference between.")
    parser.add_argument('-step', '--step',
                        action='store', default=1.0, type=float,
                        help="Initial size alpha")
    parser.add_argument("-gr", "--growth-rate",
                        action='store', default=1.0, type=float,
                        help="Growth rate of the step size alpha")
    parser.add_argument('-no-vis', '--no-visualization',
                        action='store_true',
                        default=False,
                        help='Disable or disable GLVis visualization')

    args = parser.parse_args()
    if myid == 0:
        parser.print_options(args)

    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))
    visualization = not args.no_visualization
    in_alpha = args.step

    run(meshfile=meshfile,
        order=args.order,
        max_it=args.max_it,
        ref_levels=args.refs,
        alpha=in_alpha,
        growth_rate=args.growth_rate,
        newton_scaling=0.8,
        eps=1e-6,
        tol=args.tol,
        visualization=visualization)



'''
   PyMFEM example 38

   See c++ version in the MFEM library for more detail

   Sample runs:
       python ex38.py
       python ex38.py -i volumetric1d
       python ex38.py -i surface2d
       python ex38.py -i surface2d -o 4 -r 5
       python ex38.py -i volumetric2d
       python ex38.py -i volumetric2d -o 4 -r 5
       python ex38.py -i surface3d
       python ex38.py -i surface3d -o 4 -r 5
       python ex38.py -i volumetric3d
       python ex38.py -i volumetric3d -o 4 -r 5

'''
import mfem.ser as mfem
from mfem.ser import intArray, doubleArray

import os
from os.path import expanduser, join
import numpy as np
from numpy import sin, cos, array, pi, sqrt, floor, nan

import numpy as np
from numba import njit
from numba.types import float64


class IntegrationType:
    Volumetric1D = 0
    Surface2D = 1
    Volumetric2D = 2
    Surface3D = 3
    Volumetric3D = 4


visualization = True
itype = None


def make_lvlset_coeff():
    if itype == IntegrationType.Volumetric1D:
        @mfem.jit.scalar
        def func(X):
            return .55 - X[0]
    elif itype == IntegrationType.Surface2D:
        @mfem.jit.scalar
        def func(X):
            return 1. - (X[0]**2 + X[1]**2)
    elif itype == IntegrationType.Volumetric2D:
        @mfem.jit.scalar
        def func(X):
            return 1. - ((X[0]/1.5)**2 + (X[1]/.75)**2)
    elif itype == IntegrationType.Surface3D:
        @mfem.jit.scalar
        def func(X):
            return 1. - (X[0]**2 + X[1]**2 + X[2]**2)
    elif itype == IntegrationType.Volumetric3D:
        @mfem.jit.scalar
        def func(X):
            return 1. - ((X[0]/1.5)**2 + (X[1]/0.75)**2 + (X[2]/.5)**2)
    else:
        @njit(float64(float64[:]))
        def func(X):
            return 1.
    return func


def make_integrand_coeff():
    if itype == IntegrationType.Volumetric1D:
        @mfem.jit.scalar
        def func(X):
            return 1.
    elif itype == IntegrationType.Surface2D:
        @mfem.jit.scalar
        def func(X):
            return 3. * X[0]**2 - X[1]**2
    elif itype == IntegrationType.Volumetric2D:
        @mfem.jit.scalar
        def func(X):
            return 1.
    elif itype == IntegrationType.Surface3D:
        @mfem.jit.scalar
        def func(X):
            return 4. - 3. * X[0]**2 + 2. * X[1]**2 - X[2]**2
    elif itype == IntegrationType.Volumetric3D:
        @mfem.jit.scalar
        def func(X):
            return 1.
    else:
        @mfem.jit.scalar
        def func(X):
            return 0.
    return func


def Surface():
    if itype == IntegrationType.Volumetric1D:
        return 1.
    elif itype == IntegrationType.Surface2D:
        return 2. * pi
    elif itype == IntegrationType.Volumetric2D:
        return 7.26633616541076
    elif itype == IntegrationType.Surface3D:
        return 40. / 3. * pi
    elif itype == IntegrationType.Volumetric3D:
        return 9.90182151329315
    else:
        return 0.


def Volume():
    if itype == IntegrationType.Volumetric1D:
        return .55
    elif itype == IntegrationType.Surface2D:
        return nan
    elif itype == IntegrationType.Volumetric2D:
        return 9/8*pi
    elif itype == IntegrationType.Surface3D:
        return nan
    elif itype == IntegrationType.Volumetric3D:
        return 3/4*pi
    else:
        return 0.


class SIntegrationRule(mfem.PyIntegrationRule):
    def __init__(self, Order, LvlSet, lsOrder, mesh):
        super(SIntegrationRule, self).__init__()

        self.Weights = mfem.DenseMatrix()
        self.SurfaceWeights = mfem.DenseMatrix()
        self.dim = mesh.Dimension()

        MFIRs = mfem.MomentFittingIntRules(Order, LvlSet, lsOrder)
        Tr = mesh.GetElementTransformation(0)

        ir = mfem.IntegrationRule()
        MFIRs.GetSurfaceIntegrationRule(Tr, ir)
        if self.dim > 1:
            self.Weights.SetSize(ir.GetNPoints(), mesh.GetNE())
        else:
            self.Weights.SetSize(2, mesh.GetNE())

        self.SurfaceWeights.SetSize(ir.GetNPoints(), mesh.GetNE())

        w = mfem.Vector()
        MFIRs.GetSurfaceWeights(Tr, ir, w)
        self.SurfaceWeights.SetCol(0, w)

        self.SetSize(ir.GetNPoints())

        for ip in range(self.GetNPoints()):
            self.IntPoint(ip).index = ip
            intp = self.IntPoint(ip)
            intp.x = ir.IntPoint(ip).x
            intp.y = ir.IntPoint(ip).y
            intp.z = ir.IntPoint(ip).z
            if self.dim > 1:
                self.Weights[ip, 0] = ir.IntPoint(ip).weight
            else:
                self.Weights[0, 0] = ir.IntPoint(ip).x
                self.Weights[1, 0] = ir.IntPoint(ip).weight

        for elem in range(mesh.GetNE()):
            Tr = mesh.GetElementTransformation(elem)
            MFIRs.GetSurfaceIntegrationRule(Tr, ir)

            w = mfem.Vector()
            MFIRs.GetSurfaceWeights(Tr, ir, w)
            self.SurfaceWeights.SetCol(elem, w)

            for ip in range(self.GetNPoints()):
                if self.dim > 1:
                    self.Weights[ip, elem] = ir.IntPoint(ip).weight
                else:
                    self.Weights[0, elem] = ir.IntPoint(ip).x
                    self.Weights[1, elem] = ir.IntPoint(ip).weight

    def SetElementinclSurfaceWeight(self, Element):
        if self.dim == 1:
            intp = self.IntPoint(0)
            intp.x = self.Weights[0, Element]
            intp.weight = self.Weights[1, Element]
            print(str(intp.x) + " " + str(Element))
        else:
            for ip in range(self.GetNPoints()):
                intp = self.IntPoint(ip)
                intp.weight = self.Weights[ip, Element] * \
                    self.SurfaceWeights[ip, Element]

    def SetElement(self, Element):
        if self.dim == 1:
            intp = IntPoint(0)
            intp.x = self.Weights[0, Element]
            intp.weight = Weights[1, Element]
        else:
            for ip in range(self.GetNPoints()):
                intp = self.IntPoint(ip)
                intp.weight = self.Weights[ip, Element]


class CIntegrationRule(mfem.PyIntegrationRule):
    def __init__(self, Order, LvlSet, lsOrder, mesh):
        super(CIntegrationRule, self).__init__()
        self.dim = mesh.Dimension()
        self.Weights = mfem.DenseMatrix()

        MFIRs = mfem.MomentFittingIntRules(Order, LvlSet, lsOrder)

        Tr = mesh.GetElementTransformation(0)

        ir = mfem.IntegrationRule()
        MFIRs.GetVolumeIntegrationRule(Tr, ir)
        if self.dim > 1:
            self.Weights.SetSize(ir.GetNPoints(), mesh.GetNE())
        else:
            self.Weights.SetSize(2 * ir.GetNPoints(), mesh.GetNE())

        self.SetSize(ir.GetNPoints())
        for ip in range(self.GetNPoints()):
            self.IntPoint(ip).index = ip
            intp = self.IntPoint(ip)
            intp.x = ir.IntPoint(ip).x
            intp.y = ir.IntPoint(ip).y
            intp.z = ir.IntPoint(ip).z
            if self.dim > 1:
                self.Weights[ip, 0] = ir.IntPoint(ip).weight
            else:
                self.Weights[2 * ip, 0] = ir.IntPoint(ip).x
                self.Weights[2 * ip + 1, 0] = ir.IntPoint(ip).weight

        for elem in range(mesh.GetNE()):
            Tr = mesh.GetElementTransformation(elem)
            MFIRs.GetVolumeIntegrationRule(Tr, ir)
            for ip in range(self.GetNPoints()):
                if self.dim > 1:
                    self.Weights[ip, elem] = ir.IntPoint(ip).weight
                else:
                    self.Weights[2 * ip, elem] = ir.IntPoint(ip).x
                    self.Weights[2 * ip + 1, elem] = ir.IntPoint(ip).weight

    def SetElement(self, Element):
        if self.dim == 1:
            for ip in range(self.GetNPoints()):
                intp = self.IntPoint(ip)
                intp.x = self.Weights[2 * ip, Element]
                intp.weight = self.Weights[2 * ip + 1, Element]
        else:
            for ip in range(self.GetNPoints()):
                intp = self.IntPoint(ip)
                intp.weight = self.Weights[ip, Element]


class SurfaceLFIntegrator(mfem.PyLinearFormIntegrator):
    def __init__(self, q, levelset, ir):
        self.Q = q
        self.LevelSet = levelset
        self.SIntRule = ir
        self.shape = mfem.Vector()
        super(SurfaceLFIntegrator, self).__init__(ir)

    def AssembleRHSElementVect(self, el, Tr, elvect):
        dof = el.GetDof()
        self.shape.SetSize(dof)
        elvect.SetSize(dof)
        elvect.Assign(0.)

        # Update the surface integration rule for the current element
        self.SIntRule.SetElementinclSurfaceWeight(Tr.ElementNo)

        for ip in range(self.SIntRule.GetNPoints()):
            Tr.SetIntPoint(self.SIntRule.IntPoint(ip))
            val = Tr.Weight() * self.Q.Eval(Tr, self.SIntRule.IntPoint(ip))
            el.CalcShape(self.SIntRule.IntPoint(ip), self.shape)
            mfem.add_vector(elvect, self.SIntRule.IntPoint(
                ip).weight * val, self.shape, elvect)


class SubdomainLFIntegrator(mfem.PyLinearFormIntegrator):
    def __init__(self, q, levelset, ir):
        self.shape = mfem.Vector()
        self.CIntRule = ir
        self.LevelSet = levelset
        self.Q = q
        super(SubdomainLFIntegrator, self).__init__(ir)

    def AssembleRHSElementVect(self, el, Tr, elvect):
        dof = el.GetDof()
        self.shape.SetSize(dof)
        elvect.SetSize(dof)
        elvect.Assign(0.)

        # Update the subdomain integration rule
        self.CIntRule.SetElement(Tr.ElementNo)
        for ip in range(self.CIntRule.GetNPoints()):
            Tr.SetIntPoint(self.CIntRule.IntPoint(ip))
            val = Tr.Weight() * self.Q.Eval(Tr, self.CIntRule.IntPoint(ip))
            el.CalcPhysShape(Tr, self.shape)
            mfem.add_vector(elvect, self.CIntRule.IntPoint(
                ip).weight * val, self.shape, elvect)


def run(ref_levels=3, order=2):
    if itype == IntegrationType.Volumetric1D:
        mesh = mfem.Mesh("../data/inline-segment.mesh")
    elif (itype == IntegrationType.Surface2D or
          itype == IntegrationType.Volumetric2D):
        mesh = mfem.Mesh(2, 4, 1, 0, 2)
        mesh.AddVertex(-1.6, -1.6)
        mesh.AddVertex(1.6, -1.6)
        mesh.AddVertex(1.6, 1.6)
        mesh.AddVertex(-1.6, 1.6)
        mesh.AddQuad(0, 1, 2, 3)
        mesh.FinalizeQuadMesh(1, 0, True)
    elif (itype == IntegrationType.Surface3D or
          itype == IntegrationType.Volumetric3D):
        mesh = mfem.Mesh(3, 8, 1, 0, 3)
        mesh.AddVertex(-1.6, -1.6, -1.6)
        mesh.AddVertex(1.6, -1.6, -1.6)
        mesh.AddVertex(1.6, 1.6, -1.6)
        mesh.AddVertex(-1.6, 1.6, -1.6)
        mesh.AddVertex(-1.6, -1.6, 1.6)
        mesh.AddVertex(1.6, -1.6, 1.6)
        mesh.AddVertex(1.6, 1.6, 1.6)
        mesh.AddVertex(-1.6, 1.6, 1.6)
        mesh.AddHex(0, 1, 2, 3, 4, 5, 6, 7)
        mesh.FinalizeHexMesh(1, 0, True)
    else:
        assert False, "Unknown integration type"

    for i in range(ref_levels):
        mesh.UniformRefinement()

    # 3. Define the necessary finite element space on the mesh.
    fe_col = mfem.H1_FECollection(1, mesh.Dimension())
    fespace = mfem.FiniteElementSpace(mesh, fe_col)

    # 4. Construction Coefficients for the level set and the integrand.
    levelset = make_lvlset_coeff()
    u = make_integrand_coeff()

    # 5. Define the necessary Integration rules on element 0.
    Tr = mesh.GetElementTransformation(0)

    sir = SIntegrationRule(order, levelset, 2, mesh)
    if (itype == IntegrationType.Volumetric1D or
        itype == IntegrationType.Volumetric2D or
            itype == IntegrationType.Volumetric3D):
        cir = CIntegrationRule(order, levelset, 2, mesh)

    # 6. Define and assemble the linear forms on the finite element space.
    surface = mfem.LinearForm(fespace)
    volume = mfem.LinearForm(fespace)

    surface.AddDomainIntegrator(SurfaceLFIntegrator(u, levelset, sir))
    surface.Assemble()

    if (itype == IntegrationType.Volumetric1D or
        itype == IntegrationType.Volumetric2D or
            itype == IntegrationType.Volumetric3D):
        volume.AddDomainIntegrator(SubdomainLFIntegrator(u, levelset, cir))
        volume.Assemble()

    # 7. Print information, computed values and errors to the console.
    qorder = 0
    nbasis = 2 * (order + 1) + (order * (order + 1) // 2)

    irs = mfem.IntegrationRules(0, mfem.Quadrature1D.GaussLegendre)
    ir = irs.Get(mfem.Geometry.SQUARE, qorder)

    while (ir.GetNPoints() <= nbasis):
        ir = irs.Get(mfem.Geometry.SQUARE, qorder)
        qorder += 1

    print("============================================")
    if itype != IntegrationType.Volumetric1D:
        print("Mesh size dx:                       " +
              "{:g}".format(3.2 / 2**ref_levels))
    else:
        print("Mesh size dx:                       " +
              "{:g}".format(25 / 2**ref_levels))

    if (itype == IntegrationType.Surface2D or
            itype == IntegrationType.Volumetric2D):
        print("Number of div free basis functions: " + str(nbasis))
        print("Number of quadrature points:        " + str(ir.GetNPoints()))

    print("============================================")
    print("Computed value of surface integral: " +
          "{:.2e}".format(surface.Sum()))
    print("True value of surface integral:     " + "{:.2e}".format(Surface()))

    print("Absolut Error (Surface):            " +
          "{:.2e}".format(abs(surface.Sum() - Surface())))
    print("Relative Error (Surface):           " +
          "{:.2e}".format(abs(surface.Sum() - Surface()) / Surface()))

    if (itype == IntegrationType.Volumetric1D or
        itype == IntegrationType.Volumetric2D or
            itype == IntegrationType.Volumetric3D):

        print("--------------------------------------------")
        print("Computed value of volume integral:  " +
              "{:.2e}".format(volume.Sum()))
        print("True value of volume integral:      " +
              "{:.2e}".format(Volume()))
        print("Absolut Error (Volume):             " +
              "{:.2e}".format(abs(volume.Sum() - Volume())))
        print("Relative Error (Volume):            " +
              "{:.2e}".format(abs(volume.Sum() - Volume()) / Volume()))

    print("============================================")

    # 8. Plot the level-set function on a high order finite element space.
    if visualization:
        fe_coll2 = mfem.H1_FECollection(5, mesh.Dimension())
        fespace2 = mfem.FiniteElementSpace(mesh, fe_coll2)
        levelset_coeff = make_lvlset_coeff()

        lgf = mfem.GridFunction(fespace2)
        lgf.ProjectCoefficient(levelset_coeff)

        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock.precision(8)
        sol_sock << "solution\n" << mesh << lgf
        sol_sock.flush()
        sol_sock << "keys pppppppppppppppppppppppppppcmmlRj\n"
        sol_sock << "levellines " << "0." << " " << "0." << " " << 1 << "\n"
        sol_sock.flush()


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(
        description='Ex38 (Cut-Volume and Cut-Surface Integration)')

    parser.add_argument("-o", "--order",
                        action='store', default=2, type=int,
                        help="Order (degree) of the finite elements.")
    parser.add_argument("-r", "--refine",
                        action='store', default=3, type=int,
                        help="Number of meh refinements")
    parser.add_argument("-i", "--integrationtype",
                        action='store', default="surface2d", type=str,
                        help="IntegrationType to demonstrate")
    parser.add_argument("-no-vis", "--no-visualization",
                        action='store_true', default=False,
                        help='Enable GLVis visualization')

    args = parser.parse_args()
    parser.print_options(args)

    globals()["visualization"] = not args.no_visualization

    if args.integrationtype.capitalize().startswith("Volumetric1"):
        globals()["itype"] = IntegrationType.Volumetric1D
    elif args.integrationtype.capitalize().startswith("Volumetric2"):
        globals()["itype"] = IntegrationType.Volumetric2D
    elif args.integrationtype.capitalize().startswith("Volumetric3"):
        globals()["itype"] = IntegrationType.Volumetric3D
    elif args.integrationtype.capitalize().startswith("Surface2"):
        globals()["itype"] = IntegrationType.Surface2D
    elif args.integrationtype.capitalize().startswith("Surface3"):
        globals()["itype"] = IntegrationType.Surface3D
    else:
        assert False, "Unknown integration type"

    if not hasattr(mfem, "MomentFittingIntRules"):
        assert False, "MFEM is not built with Lapack"

    run(ref_levels=args.refine,
        order=args.order)




'''
   MFEM example 24
      See c++ version in the MFEM library for more detail 
'''
import os
import mfem.ser as mfem
from mfem.ser import intArray
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sin, cos, exp, sqrt, pi, abs, array, floor, log

freq = 1.0


def run(order=1,
        prob=0,
        static_cond=False,
        mesh_file="",
        visualization=1,
        device="cpu",
        pa=False,
        numba=True):

    kappa = freq * pi

    # 2. Enable hardware devices such as GPUs, and programming models such as
    #    CUDA, OCCA, RAJA and OpenMP based on command line options.
    device = mfem.Device(device)
    device.Print()

    # 3. Read the mesh from the given mesh file. We can handle triangular,
    #    quadrilateral, tetrahedral, hexahedral, surface and volume meshes with
    #    the same code.
    mesh = mfem.Mesh(mesh_file, 1, 1)
    dim = mesh.Dimension()
    sdim = mesh.SpaceDimension()

    # 4. Refine the mesh to increase the resolution. In this example we do
    #    'ref_levels' of uniform refinement. We choose 'ref_levels' to be the
    #    largest number that gives a final mesh with no more than 50,000
    #    elements.
    ref_levels = int(floor(log(50000./mesh.GetNE())/log(2.)/dim))
    for l in range(ref_levels):
        mesh.UniformRefinement()
    mesh.ReorientTetMesh()

    # 5. Define a finite element space on the mesh. Here we use Nedelec or
    #   Raviart-Thomas finite elements of the specified order.
    if prob == 0:
        trial_fec = mfem.H1_FECollection(order, dim)
        test_fec = mfem.ND_FECollection(order, dim)
    elif prob == 1:
        trial_fec = mfem.ND_FECollection(order, dim)
        test_fec = mfem.RT_FECollection(order-1, dim)
    else:
        trial_fec = mfem.RT_FECollection(order-1, dim)
        test_fec = mfem.L2_FECollection(order-1, dim)

    trial_fes = mfem.FiniteElementSpace(mesh, trial_fec)
    test_fes = mfem.FiniteElementSpace(mesh, test_fec)

    trial_size = trial_fes.GetTrueVSize()
    test_size = test_fes.GetTrueVSize()

    if prob == 0:
        print("Number of Nedelec finite element unknowns: " + str(test_size))
        print("Number of H1 finite element unknowns: " + str(trial_size))
    elif prob == 1:
        print("Number of Nedelec finite element unknowns: " + str(test_size))
        print("Number of Raviart-Thomas finite element unknowns: " + str(trial_size))
    else:
        print("Number of Raviart-Thomas finite element unknowns: " + str(test_size))
        print("Number of L2 finite element unknowns: " + str(trial_size))

    # 6. Define the solution vector as a finite element grid function
    #    corresponding to the trial fespace.
    gftest = mfem.GridFunction(test_fes)
    gftrial = mfem.GridFunction(trial_fes)
    x = mfem.GridFunction(test_fes)

    if numba:
        @mfem.jit.scalar()
        def p_coef(x):
            if dim == 3:
                return sin(x[0]) * sin(x[1]) * sin(x[2])
            elif dim == 2:
                return sin(x[0]) * sin(x[1])
            return 0.0

        @mfem.jit.vector(shape=(sdim,))
        def gradp_coef(x):
            out = np.zeros(shape, dtype=np.float64)
            if dim == 3:
                out[0] = cos(x[0]) * sin(x[1]) * sin(x[2])
                out[1] = sin(x[0]) * cos(x[1]) * sin(x[2])
                out[2] = sin(x[0]) * sin(x[1]) * cos(x[2])
            elif dim == 2:
                out[0] = cos(x[0]) * sin(x[1])
                out[1] = sin(x[0]) * cos(x[1])
                if len(x) == 3:
                    out[2] = 0.0
            return out

        @mfem.jit.vector(vdim=sdim, interface="c++")
        def v_coef(x, out):
            out = np.zeros(shape, dtype=np.float64)
            if dim == 3:
                out[0] = sin(kappa * x[1])
                out[1] = sin(kappa * x[2])
                out[2] = sin(kappa * x[0])
            else:
                out[0] = sin(kappa * x[1])
                out[1] = sin(kappa * x[0])
                if shape[0] == 3:
                    out[2] = 0.0

        @mfem.jit.scalar()
        def cdiv_gradp_coeff(x):
            if dim == 3:
                return -3.0 * sin(x[0]) * sin(x[1]) * sin(x[2])
            elif dim == 2:
                return -2.0 * sin(x[0]) * sin(x[1])
            return 0.0

        @mfem.jit.vector(shape=(sdim,), interface="c++")
        def curlv_coef(x, out):
            if dim == 3:
                out[0] = -kappa * cos(kappa * x[2])
                out[1] = -kappa * cos(kappa * x[0])
                out[2] = -kappa * cos(kappa * x[1])
            else:
                for i in range(len(out)):
                    out[i] = 0.0

        @mfem.jit.scalar()
        def divgradp_coef(x):
            if dim == 3:
                return -3.0 * sin(x[0]) * sin(x[1]) * sin(x[2])
            elif dim == 2:
                return -2.0 * sin(x[0]) * sin(x[1])
            return 0.0

    else:
        class cp_exact(mfem.PyCoefficient):
            def EvalValue(self, x):
                if dim == 3:
                    return sin(x[0]) * sin(x[1]) * sin(x[2])
                elif dim == 2:
                    return sin(x[0]) * sin(x[1])
                return 0.0

        class cgradp_exact(mfem.VectorPyCoefficient):
            def __init__(self):
                mfem.VectorPyCoefficient.__init__(self, sdim)

            def EvalValue(self, x):
                if dim == 3:
                    return [cos(x[0]) * sin(x[1]) * sin(x[2]),
                            sin(x[0]) * cos(x[1]) * sin(x[2]),
                            sin(x[0]) * sin(x[1]) * cos(x[2]), ]
                elif dim == 2:
                    v1 = cos(x[0]) * sin(x[1])
                    v2 = sin(x[0]) * cos(x[1])
                    if len(x) == 3:
                        return v1, v2, 0.0
                    else:
                        return v1, v2

        class cdiv_gradp_exact(mfem.PyCoefficient):
            def EvalValue(self, x):
                if dim == 3:
                    return -3.0 * sin(x[0]) * sin(x[1]) * sin(x[2])
                elif dim == 2:
                    return -2.0 * sin(x[0]) * sin(x[1])
                return 0.0

        class cv_exact(mfem.VectorPyCoefficient):
            def __init__(self):
                mfem.VectorPyCoefficient.__init__(self, sdim)

            def EvalValue(self, x):
                if dim == 3:
                    return [sin(kappa * x[1]),
                            sin(kappa * x[2]),
                            sin(kappa * x[0])]
                else:
                    v1 = sin(kappa * x[1])
                    v2 = sin(kappa * x[0])
                    if len(x) == 3:
                        return (v1, v2, 0.0)
                    else:
                        return (v1, v2)

        class ccurlv_exact(mfem.PyCoefficient):
            def EvalValue(self, x):
                if dim == 3:
                    return [-kappa * cos(kappa * x[2]),
                            -kappa * cos(kappa * x[0]),
                            -kappa * cos(kappa * x[1]), ]
                else:
                    return [0.0] * self.vdim

        p_coef = cp_exact()
        gradp_coef = cgradp_exact()
        v_coef = cv_exact()
        curlv_coef = ccurlv_exact()
        divgradp_coef = cdiv_gradp_exact()

    if prob == 0:
        gftrial.ProjectCoefficient(p_coef)
    elif prob == 1:
        gftrial.ProjectCoefficient(v_coef)
    else:
        gftrial.ProjectCoefficient(gradp_coef)

    gftrial.SetTrueVector()
    gftrial.SetFromTrueVector()

    # 7. Set up the bilinear forms for L2 projection.
    one = mfem.ConstantCoefficient(1.0)
    a = mfem.BilinearForm(test_fes)
    a_mixed = mfem.MixedBilinearForm(trial_fes, test_fes)
    if pa:
        a.SetAssemblyLevel(mfem.AssemblyLevel_PARTIAL)
        a_mixed.SetAssemblyLevel(mfem.AssemblyLevel_PARTIAL)

    if prob == 0:
        a.AddDomainIntegrator(mfem.VectorFEMassIntegrator(one))
        a_mixed.AddDomainIntegrator(mfem.MixedVectorGradientIntegrator(one))
    elif prob == 1:
        a.AddDomainIntegrator(mfem.VectorFEMassIntegrator(one))
        a_mixed.AddDomainIntegrator(mfem.MixedVectorCurlIntegrator(one))
    else:
        a.AddDomainIntegrator(mfem.MassIntegrator(one))
        a_mixed.AddDomainIntegrator(mfem.VectorFEDivergenceIntegrator(one))

    # 8. Assemble the bilinear form and the corresponding linear system,
    #    applying any necessary transformations such as: eliminating boundary
    #    conditions, applying conforming constraints for non-conforming AMR,
    #    static condensation, etc.
    if static_cond:
        a.EnableStaticCondensation()

    a.Assemble()
    if not pa:
        a.Finalize()

    a_mixed.Assemble()
    if not pa:
        a_mixed.Finalize()

    if pa:
        a_mixed.Mult(gftrial, x)
    else:
        mixed = a_mixed.SpMat()
        mixed.Mult(gftrial, x)

    # 9. Define and apply a PCG solver for Ax = b with Jacobi preconditioner.
    rhs = mfem.GridFunction(test_fes)
    rhs.Assign(x)
    x.Assign(0.0)

    cg = mfem.CGSolver()
    cg.SetRelTol(1e-12)
    cg.SetMaxIter(1000)
    cg.SetPrintLevel(1)
    if pa:
        ess_tdof_list = mfem.intArray()
        Jacobi = mfem.OperatorJacobiSmoother(a, ess_tdof_list)
        cg.SetOperator(a)
        cg.SetPreconditioner(Jacobi)
        cg.Mult(rhs, x)
    else:
        Amat = a.SpMat()
        Jacobi = mfem.DSmoother(Amat)
        cg.SetOperator(Amat)
        cg.SetPreconditioner(Jacobi)
        cg.Mult(rhs, x)

    # 10. Compute the same field by applying a DiscreteInterpolator.
    discreteInterpolant = mfem.GridFunction(test_fes)
    dlo = mfem.DiscreteLinearOperator(trial_fes, test_fes)
    if prob == 0:
        dlo.AddDomainInterpolator(mfem.GradientInterpolator())
    elif prob == 1:
        dlo.AddDomainInterpolator(mfem.CurlInterpolator())
    else:
        dlo.AddDomainInterpolator(mfem.DivergenceInterpolator())

    dlo.Assemble()
    dlo.Mult(gftrial, discreteInterpolant)

    # 11. Compute the projection of the exact field.
    exact_proj = mfem.GridFunction(test_fes)
    if prob == 0:
        exact_proj.ProjectCoefficient(gradp_coef)
    elif prob == 1:
        exact_proj.ProjectCoefficient(curlv_coef)
    else:
        exact_proj.ProjectCoefficient(divgradp_coef)

    exact_proj.SetTrueVector()
    exact_proj.SetFromTrueVector()

    print()  # put empty line
    # 12. Compute and print the L_2 norm of the error.
    if prob == 0:
        errSol = x.ComputeL2Error(gradp_coef)
        errInterp = discreteInterpolant.ComputeL2Error(gradp_coef)
        errProj = exact_proj.ComputeL2Error(gradp_coef)

        print(" Solution of (E_h,v) = (grad p_h,v) for E_h and v in H(curl): " +
              "|| E_h - grad p ||_{L_2} = " + "{:g}".format(errSol) + "\n")
        print(" Gradient interpolant E_h = grad p_h in H(curl): || E_h - grad p" +
              " ||_{L_2} = " + "{:g}".format(errInterp) + "\n")
        print(" Projection E_h of exact grad p in H(curl): || E_h - grad p " +
              "||_{L_2} = " + "{:g}".format(errProj) + "\n")
    elif prob == 1:
        errSol = x.ComputeL2Error(curlv_coef)
        errInterp = discreteInterpolant.ComputeL2Error(curlv_coef)
        errProj = exact_proj.ComputeL2Error(curlv_coef)

        print(" Solution of (E_h,w) = (curl v_h,w) for E_h and w in H(div): " +
              "|| E_h - curl v ||_{L_2} = " + "{:g}".format(errSol) + "\n")
        print(" Curl interpolant E_h = curl v_h in H(div): || E_h - curl v " +
              "||_{L_2} = " + "{:g}".format(errInterp) + "\n")
        print(" Projection E_h of exact curl v in H(div): || E_h - curl v " +
              "||_{L_2} = " + "{:g}".format(errProj) + "\n")
    else:
        order_quad = max(2, 2*order+1)
        irs = [mfem.IntRules.Get(i, order_quad)
               for i in range(mfem.Geometry.NumGeom)]

        errSol = x.ComputeL2Error(divgradp_coef, irs)
        errInterp = discreteInterpolant.ComputeL2Error(divgradp_coef, irs)
        errProj = exact_proj.ComputeL2Error(divgradp_coef, irs)

        print(" Solution of (f_h,q) = (div v_h,q) for f_h and q in L_2: " +
              "|| f_h - div v ||_{L_2} = " + "{:g}".format(errSol) + "\n")
        print(" Divergence interpolant f_h = div v_h in L_2: || f_h - div v " +
              "||_{L_2} = " + "{:g}".format(errInterp) + "\n")
        print(" Projection f_h of exact div v in L_2: || f_h - div v " +
              "||_{L_2} = " + "{:g}".format(errProj) + "\n")

    # 13. Save the refined mesh and the solution. This output can be viewed
    #     later using GLVis: "glvis -m refined.mesh -g sol.gf".
    mesh.Print("refined.mesh", 8)
    x.Save("sol.gf", 8)

    # 14. Send the solution by socket to a GLVis server.
    if visualization:
        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock.precision(8)
        sol_sock << "solution\n" << mesh << x
        sol_sock.flush()


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex24 (Mixed finite element spaces)')

    parser.add_argument('-m', '--mesh',
                        default='beam-hex.mesh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-p', '--problem-type',
                        action='store', type=int, default=0,
                        help="Choose between 0: grad, 1: curl, 2: div")
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree)")
    parser.add_argument('-sc', '--static-condensation',
                        action='store_true',
                        help="Enable static condensation.")
    parser.add_argument("-pa", "--partial-assembly",
                        action='store_true',
                        help="Enable Partial Assembly.")
    parser.add_argument("-d", "--device",
                        default="cpu", type=str,
                        help="Device configuration string, see Device::Configure().")
    try:
        from numba import jit
        HAS_NUMBA = True
    except ImportError:
        HAS_NUMBA = False
    parser.add_argument("-n", "--numba",
                        default=int(HAS_NUMBA),
                        type=int,
                        help="Use Number compiled coefficient")

    args = parser.parse_args()
    args.numba = bool(args.numba)
    parser.print_options(args)

    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))
    numba = (args.numba == 1)

    run(order=args.order,
        prob=args.problem_type,
        static_cond=args.static_condensation,
        mesh_file=meshfile,
        visualization=args.visualization,
        device=args.device,
        pa=args.partial_assembly,
        numba=numba)



import os
path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

mfem_mode = None
pymfem_debug = -1

def debug_print(message):
    if pymfem_debug < 0:   # debug < 0 
        return
    elif pymfem_debug == 0:   # debug = 0
        pass
    elif pymfem_debug > 0:   # debug = 1
        pass
    elif pymfem_debug > 1: # debug = 2
        import traceback as tb
        tb.print_stack()
    elif pymfem_debug > 2: # debug = 3
        import traceback as tb
        print(''.join(tb.format_list([tb.extract_stack(limit = 2)[0],])))

    print(message)

__version__ = '4.8.0.1'




"""
  MFEM + PyMFEM (finite element method library)
"""
import sys
import os
import site
import re
import shutil
import subprocess
from shutil import which as find_command

from sys import platform
import multiprocessing
from multiprocessing import Pool
# force fork instead of spawn on MacOS to avoid race condition on mfem/__init__.py
if platform == "darwin":
    multiprocessing.set_start_method("fork")

from setuptools import setup, find_packages
from setuptools.command.build_py import build_py as _build_py
from setuptools.command.install import install as _install
from setuptools.command.install_egg_info import install_egg_info as _install_egg_info
from setuptools.command.install_lib import install_lib as _install_lib
from setuptools.command.install_scripts import install_scripts as _install_scripts
from setuptools.command.bdist_wheel import bdist_wheel as _bdist_wheel
from distutils.command.clean import clean as _clean

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "_build_system"))

import build_globals as bglb
from build_mfem import *
from build_metis import *
from build_hypre import *
from build_pymfem import *
from build_libceed import *
from build_gslib import *
from build_config import *
from build_consts import *
from build_utils import *

# ----------------------------------------------------------------------------------------
# Constants
# ----------------------------------------------------------------------------------------

class Install(_install):
    '''
    called when pyton setup.py install
    '''
    user_options = _install.user_options + cmd_options

    def initialize_options(self):
        _install.initialize_options(self)
        initialize_cmd_options(self)

    def finalize_options(self):
        if (bool(self.ext_only) and bool(self.skip_ext)):
            assert False, "skip-ext and ext-only can not use together"

        given_prefix = True
        if (self.prefix == '' or
                self.prefix is None):
            given_prefix = False
        else:
            self.prefix = os.path.expanduser(self.prefix)
            prefix = self.prefix

        bglb.verbose = bool(self.vv)
        if given_prefix:
            # global ext_prefix
            self.prefix = abspath(bglb.prefix)
            # ext_prefix = abspath(prefix)
        else:
            if '--user' in sys.argv:
                path = site.getusersitepackages()
                if not os.path.exists(path):
                    try:
                        print("!!!!! attempting to make a --user directory", path)
                        os.makedirs(path)
                    except BaseException:
                        pass
                if os.path.exists(path):
                    path = os.path.dirname(path)
                    path = os.path.dirname(path)
                    path = os.path.dirname(path)
                else:
                    assert False, "no installation path found"
                self.prefix = path
            else:
                self.prefix = sys.prefix

        self.user = 0
        _install.finalize_options(self)

        bglb.use_unverifed_SSL = self.unverifiedSSL

        if bglb.verbose:
            print("!!!!! prefix is :", self.prefix)

    def run(self):
        if not bglb.is_configured:
            if bglb.verbose:
                 print('!!!!! Running config (install)')
            bglb.prefix = abspath(self.prefix)
            configure_build(self)
            print_config()

        if bglb.swig_only and not bglb.build_py_done:
            #  comes here if python setup.py install is used
            self.run_command("build")
        else:
            _install.run(self)


class BdistWheel(_bdist_wheel):
    '''
    Wheel build performs SWIG + Serial in Default.
    --skip-build option skip building entirely.
    '''
    user_options = _bdist_wheel.user_options + cmd_options

    def initialize_options(self):
        _bdist_wheel.initialize_options(self)
        initialize_cmd_options(self)

    def finalize_options(self):
        def _has_ext_modules():
            return True
        self.distribution.has_ext_modules = _has_ext_modules
        _bdist_wheel.finalize_options(self)

    def run(self):
        import build_globals as bglb

        if not bglb.is_configured:
            if bglb.verbose:
                print('!!!!! Running config (bdist wheel)')
            bglb.prefix = abspath(self.bdist_dir)
            bglb.ext_prefix = os.path.join(bglb.prefix, 'mfem', 'external')
            bglb.bdist_wheel_dir = abspath(self.bdist_dir)
            bglb.do_bdist_wheel = True

            configure_build(self)
            clean_dist_info(bglb.prefix)
            if bglb.keep_temp:
                self.keep_temp = True
            print_config()

        self.run_command("build")

        _bdist_wheel.run(self)


class BuildPy(_build_py):
    '''
    Called when python setup.py build_py
    '''
    user_options = _build_py.user_options

    def initialize_options(self):
        _build_py.initialize_options(self)

    def finalize_options(self):
        _build_py.finalize_options(self)

    def run(self):
        bglb.build_py_done = True

        if not bglb.swig_only:
            if bglb.build_metis:
                if bglb.use_metis_gklib:
                    gitclone('gklib', use_sha=True)
                    gitclone('metis', use_sha=True)
                    make_metis(use_int64=bglb.metis_64,
                               use_real64=bglb.metis_64)
                else:
                    download('metis')
                    make_metis(use_int64=bglb.metis_64,
                               use_real64=bglb.metis_64)

            if bglb.build_hypre:
                download('hypre')
                cmake_make_hypre()
            if bglb.build_libceed:
                download('libceed')
                make_libceed()
            if bglb.build_gslib:
                download('gslib')
                make_gslib(serial=True)
                if bglb.build_hypre:
                    make_gslib()

            mfem_downloaded = False

            if bglb.build_mfem:
                gitclone('mfem', use_sha=True) if bglb.mfem_branch is None else gitclone(
                    'mfem', branch=bglb.mfem_branch)
                mfem_downloaded = True
                cmake_make_mfem(serial=True)

            if bglb.build_mfemp:
                if not mfem_downloaded:
                    gitclone('mfem', use_sha=True) if bglb.mfem_branch is None else gitclone(
                        'mfem', branch=bglb.mfem_branch)
                cmake_make_mfem(serial=False)

        if bglb.clean_swig:
            clean_wrapper()
        if bglb.run_swig:
            generate_wrapper(bglb.run_swig_parallel)

        if bglb.build_serial:
            make_mfem_wrapper(serial=True)
        if bglb.build_parallel:
            make_mfem_wrapper(serial=False)

        _build_py.run(self)

class InstallLib(_install_lib):
    def finalize_options(self):
        _install_lib.finalize_options(self)
        src_cmd_obj = self.distribution.get_command_obj('install')
        src_cmd_obj.ensure_finalized()
        self.install_dir = src_cmd_obj.install_platlib

    def run(self):
        if not bglb.dry_run:
            _install_lib.run(self)
        else:
            print("skipping regular install_lib")


class InstallEggInfo(_install_egg_info):
    def run(self):
        if not bglb.dry_run:
            _install_egg_info.run(self)
        else:
            print("skipping regular install_egg_info")


class InstallScripts(_install_scripts):
    def run(self):
        if not bglb.dry_run:
            _install_scripts.run(self)
        else:
            print("skipping regular install_scripts")


class Clean(_clean):
    '''
    Called when python setup.py clean
    '''
    user_options = _clean.user_options + [
        ('ext', None, 'clean exteranal dependencies)'),
        ('mfem', None, 'clean mfem'),
        ('metis', None, 'clean metis'),
        ('hypre', None, 'clean hypre'),
        ('swig', None, 'clean swig'),
        ('all-exts', None, 'delete all externals'),
    ]

    def initialize_options(self):
        _clean.initialize_options(self)
        self.ext = False
        self.mfem = False
        self.hypre = False
        self.metis = False
        self.swig = False
        self.all_exts = False

    def run(self):
        bglb.dry_run = self.dry_run
        bglb.verbose = bool(self.verbose)

        os.chdir(extdir)

        make_command = find_command('make')

        if self.ext or self.mfem:
            path = os.path.join(extdir, 'mfem', 'cmbuild_par')
            if os.path.exists(path):
                shutil.rmtree(path)
            path = os.path.join(extdir, 'mfem', 'cmbuild_ser')
            if os.path.exists(path):
                shutil.rmtree(path)
        if self.ext or self.hypre:
            path = os.path.join(extdir, 'hypre', 'cmbuild')
            if os.path.exists(path):
                shutil.rmtree(path)
        if self.ext or self.metis:
            path = os.path.join(extdir, 'metis')
            if os.path.exists(path):
                os, chdir(path)
                command = ['make', 'clean']
                subprocess.check_call(command)
        if self.all_exts or self.all:
            for xxx in ('metis', 'hypre', 'mfem', 'gslib', 'gklib', 'libceed'):
                path = os.path.join(extdir, xxx)
                if os.path.exists(path):
                    shutil.rmtree(path)

        if self.swig or self.all:
            clean_wrapper()

        clean_so(all=self.all)

        os.chdir(rootdir)
        _clean.run(self)


if __name__ == '__main__':
    cmdclass = {'build_py': BuildPy,
                'install': Install,
                'install_lib': InstallLib,
                'install_egg_info': InstallEggInfo,
                'install_scripts': InstallScripts,
                'clean': Clean,
                'bdist_wheel': BdistWheel}

    setup(
        cmdclass=cmdclass,
    )



'''
   MFEM example 18 parallel
      This is a version of Example 18 with a simple adaptive mesh
      refinement loop. 
      See c++ version in the MFEM library for more detail 
   Sample runs:

       mpirun -np 4 python ex18p.py -p 1 -rs 2 -rp 1 -o 1 -s 3
       mpirun -np 4 python ex18p.py -p 1 -rs 1 -rp 1 -o 3 -s 4
       mpirun -np 4 python ex18p.py -p 1 -rs 1 -rp 1 -o 5 -s 6
       mpirun -np 4 python ex18p.py -p 2 -rs 1 -rp 1 -o 1 -s 3 -mf
       mpirun -np 4 python ex18p.py -p 2 -rs 1 -rp 1 -o 3 -s 3 -mf

'''
import mfem.par as mfem

from mfem.common.arg_parser import ArgParser

from os.path import expanduser, join, dirname
import numpy as np
from numpy import sqrt, pi, cos, sin, hypot, arctan2
from scipy.special import erfc

from ex18_common import (EulerMesh,
                         EulerInitialCondition,
                         DGHyperbolicConservationLaws)

# 1. Initialize MPI.from mpi4py import MPI

from mpi4py import MPI
num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank


def run(problem=1,
        ser_ref_levels=0,
        par_ref_levels=1,
        order=3,
        ode_solver_type=4,
        t_final=2.0,
        dt=-0.01,
        cfl=0.3,
        visualization=True,
        vis_steps=50,
        preassembleWeakDiv=False,
        meshfile=''):

    specific_heat_ratio = 1.4
    gas_constant = 1.0
    IntOrderOffset = 1

    device = mfem.Device('cpu')
    if myid == 0:
        device.Print()

    # 2. Read the mesh from the given mesh file. When the user does not provide
    #    mesh file, use the default mesh file for the problem.

    mesh = EulerMesh(meshfile, problem)
    dim = mesh.Dimension()
    num_equation = dim + 2

    # Refine the mesh to increase the resolution. In this example we do
    # 'ser_ref_levels' of uniform refinement, where 'ser_ref_levels' is a
    # command-line parameter.
    for lev in range(ser_ref_levels):
        mesh.UniformRefinement()

    # Define a parallel mesh by a partitioning of the serial mesh. Refine this
    # mesh further in parallel to increase the resolution. Once the parallel
    # mesh is defined, the serial mesh can be deleted.
    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
    del mesh

    # Refine the mesh to increase the resolution. In this example we do
    # 'par_ref_levels' of uniform refinement, where 'par_ref_levels' is a
    # command-line parameter.
    for lev in range(par_ref_levels):
        pmesh.UniformRefinement()

    # 3. Define the ODE solver used for time integration. Several explicit
    #    Runge-Kutta methods are available.
    ode_solver = None
    if ode_solver_type == 1:
        ode_solver = mfem.ForwardEulerSolver()
    elif ode_solver_type == 2:
        ode_solver = mfem.RK2Solver(1.0)
    elif ode_solver_type == 3:
        ode_solver = mfem.RK3SSPSolver()
    elif ode_solver_type == 4:
        ode_solver = mfem.RK4Solver()
    elif ode_solver_type == 6:
        ode_solver = mfem.RK6Solver()
    else:
        print("Unknown ODE solver type: " + str(ode_solver_type))
        exit

    # 4. Define the discontinuous DG finite element space of the given
    #    polynomial order on the refined mesh.
    fec = mfem.DG_FECollection(order, dim)
    # Finite element space for a scalar (thermodynamic quantity)
    fes = mfem.ParFiniteElementSpace(pmesh, fec)
    # Finite element space for a mesh-dim vector quantity (momentum)
    dfes = mfem.ParFiniteElementSpace(pmesh, fec, dim, mfem.Ordering.byNODES)
    # Finite element space for all variables together (total thermodynamic state)
    vfes = mfem.ParFiniteElementSpace(
        pmesh, fec, num_equation, mfem.Ordering.byNODES)

    assert fes.GetOrdering() == mfem.Ordering.byNODES, "Ordering must be byNODES"
    glob_size = vfes.GlobalTrueVSize()
    if myid == 0:
        print("Number of unknowns: " + str(glob_size))

    # 5. Define the initial conditions, save the corresponding mesh and grid
    #    functions to files. These can be opened with GLVis using:
    #    "glvis -np 4  -m euler-mesh -g euler-1-init" (for x-momentum).

    # Initialize the state.
    u0 = EulerInitialCondition(problem,
                               specific_heat_ratio,
                               gas_constant)
    sol = mfem.ParGridFunction(vfes)
    sol.ProjectCoefficient(u0)

    # (Python note): GridFunction pointing to the subset of vector FES.
    #  sol is Vector with dim*fes.GetNDofs()
    #  Since sol.GetDataArray() returns numpy array pointing to the data, we make
    #  Vector from a sub-vector of the returned numpy array and pass it to GridFunction
    #  constructor.
    mom = mfem.GridFunction(dfes, mfem.Vector(
        sol.GetDataArray()[fes.GetNDofs():]))

    # Output the initial solution.
    smyid = '{:0>6d}'.format(myid)
    pmesh.Print("euler-mesh."+smyid, 8)
    for k in range(num_equation):
        uk = mfem.ParGridFunction(fes, mfem.Vector(
            sol.GetDataArray()[k*fes.GetNDofs():]))
        sol_name = "euler-" + str(k) + "-init."+smyid
        uk.Save(sol_name, 8)

    # 6. Set up the nonlinear form with euler flux and numerical flux
    flux = mfem.EulerFlux(dim, specific_heat_ratio)
    numericalFlux = mfem.RusanovFlux(flux)
    formIntegrator = mfem.HyperbolicFormIntegrator(
        numericalFlux, IntOrderOffset)

    euler = DGHyperbolicConservationLaws(vfes, formIntegrator,
                                         preassembleWeakDivergence=preassembleWeakDiv)

    # 7. Visualize momentum with its magnitude
    if (visualization):
        MPI.COMM_WORLD.Barrier()
        sout = mfem.socketstream("localhost", 19916)
        sout.precision(8)
        sout << "parallel " << str(num_procs) << " " << str(myid) << "\n"
        sout << "solution\n" << pmesh << mom
        sout << "window_title 'momentum, t = 0'\n"
        sout << "view 0 0\n"  # view from top
        sout << "keys jlm\n"  # turn off perspective and light, show mesh
        sout << "pause\n"
        sout.flush()

        if myid == 0:
            print("GLVis visualization paused.")
            print(" Press space (in the GLVis window) to resume it.")

    # 8. Time integration
    my_hmin = np.inf
    if (cfl > 0):
        my_hmin = min([pmesh.GetElementSize(i, 1)
                      for i in range(pmesh.GetNE())])

        hmin = MPI.COMM_WORLD.allreduce(my_hmin, op=MPI.MIN)

        # Find a safe dt, using a temporary vector. Calling Mult() computes the
        # maximum char speed at all quadrature points on all faces (and all
        # elements with -mf).
        z = mfem.Vector(sol.Size())
        euler.Mult(sol, z)

        my_max_char_speed = euler.GetMaxCharSpeed()
        max_char_speed = MPI.COMM_WORLD.allreduce(
            my_max_char_speed, op=MPI.MAX)

        dt = cfl * hmin / max_char_speed / (2 * order + 1)

    t = 0.0
    euler.SetTime(t)
    ode_solver.Init(euler)

    # Integrate in time.
    done = False
    ti = 0
    while not done:
        dt_real = min(dt, t_final - t)
        t, dt_real = ode_solver.Step(sol, t, dt_real)

        if (cfl > 0):
            my_max_char_speed = euler.GetMaxCharSpeed()
            max_char_speed = MPI.COMM_WORLD.allreduce(
                my_max_char_speed, op=MPI.MAX)

            dt = cfl * hmin / max_char_speed / (2*order+1)

        ti = ti+1
        done = (t >= t_final - 1e-8*dt)
        if (done or ti % vis_steps == 0):
            if myid == 0:
                print("time step: " + str(ti) + ", time: " + "{:g}".format(t))
            if (visualization):
                sout << "window_title 'momentum, t = " << "{:g}".format(
                    t) << "'\n"
                sout << "parallel " << str(
                    num_procs) << " " << str(myid) << "\n"
                sout << "solution\n" << pmesh << mom
                sout.flush()

    if myid == 0:
        print("done")

    # 9. Save the final solution. This output can be viewed later using GLVis:
    #    "glvis -np 4 -m euler-mesh-final -g euler-1-final" (for x-momentum).
    pmesh.Print("euler-mesh-final."+smyid, 8)
    for k in range(num_equation):
        uk = mfem.ParGridFunction(fes, mfem.Vector(
            sol.GetDataArray()[k*fes.GetNDofs():]))
        sol_name = "euler-" + str(k) + "-final."+smyid
        uk.Save(sol_name, 8)

    # 10. Compute the L2 solution error summed for all components.
    if True:
        error = sol.ComputeLpError(2., u0)
        if myid == 0:
            print("Solution error: " + "{:g}".format(error))


if __name__ == "__main__":

    parser = ArgParser(description='Ex18p')
    parser.add_argument('-m', '--mesh',
                        default='',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-p', '--problem',
                        action='store', default=1, type=int,
                        help='Problem setup to use. See options in velocity_function().')
    parser.add_argument('-rs', '--refine_serial',
                        action='store', default=0, type=int,
                        help="Number of times to refine the mesh uniformly before parallel.")
    parser.add_argument('-rp', '--refine_parallel',
                        action='store', default=1, type=int,
                        help="Number of times to refine the mesh uniformly after parallel.")
    parser.add_argument('-o', '--order',
                        action='store', default=3, type=int,
                        help="Finite element order (polynomial degree)")
    parser.add_argument('-s', '--ode_solver',
                        action='store', default=4, type=int,
                        help="ODE solver: 1 - Forward Euler,\n\t" +
                        "            2 - RK2 SSP, 3 - RK3 SSP, 4 - RK4, 6 - RK6.")
    parser.add_argument('-tf', '--t_final',
                        action='store', default=2.0, type=float,
                        help="Final time; start time is 0.")
    parser.add_argument("-dt", "--time_step",
                        action='store', default=-0.01, type=float,
                        help="Time step.")
    parser.add_argument('-c', '--cfl_number',
                        action='store', default=0.3, type=float,
                        help="CFL number for timestep calculation.")
    parser.add_argument('-novis', '--no_visualization',
                        action='store_true', default=False,
                        help='Disable GLVis visualization')
    parser.add_argument("-ea", "--element-assembly-divergence",
                        action='store_true', default=False,
                        help="Weak divergence assembly level\n" +
                        "    ea - Element assembly with interpolated")
    parser.add_argument("-mf", "--matrix-free-divergence",
                        action='store_true', default=False,
                        help="Weak divergence assembly level\n" +
                        "    mf - Nonlinear assembly in matrix-free manner")
    parser.add_argument('-vs', '--visualization-steps',
                        action='store', default=50, type=float,
                        help="Visualize every n-th timestep.")

    args = parser.parse_args()

    visualization = not args.no_visualization

    if (not args.matrix_free_divergence and
            not args.element_assembly_divergence):
        args.element_assembly_divergence = True
        args.matrix_free_divergence = False
        preassembleWeakDiv = True

    elif args.element_assembly_divergence:
        args.matrix_free_divergence = False
        preassembleWeakDiv = True

    elif args.matrix_free_divergence:
        args.element_assembly_divergence = False
        preassembleWeakDiv = False

    if myid == 0:
        parser.print_options(args)

    run(problem=args.problem,
        ser_ref_levels=args.refine_serial,
        par_ref_levels=args.refine_parallel,
        order=args.order,
        ode_solver_type=args.ode_solver,
        t_final=args.t_final,
        dt=args.time_step,
        cfl=args.cfl_number,
        visualization=visualization,
        vis_steps=args.visualization_steps,
        preassembleWeakDiv=preassembleWeakDiv,
        meshfile=args.mesh)



'''
   MFEM example 9
      This is a version of Example 1 with a simple adaptive mesh
      refinement loop. 
      See c++ version in the MFEM library for more detail 
'''
from mfem import path
import mfem.par as mfem
from mfem.par import intArray
from os.path import expanduser, join, dirname, exists
from mpi4py import MPI
import numpy as np
from numpy import sqrt, pi, cos, sin, hypot, arctan2
from scipy.special import erfc

num_proc = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
verbose = (myid == 0)

problem = 0
ser_ref_levels = 2
par_ref_levels = 0
order = 3
ode_solver_type = 4
t_final = 10
dt = 0.01
vis_steps = 5

device = mfem.Device('cpu')
if myid == 0:
    device.Print()

# 3. Read the serial mesh from the given mesh file on all processors. We can
#    handle geometrically periodic meshes in this code.
meshfile = expanduser(join(path, 'data', 'periodic-hexagon.mesh'))
if not exists(meshfile):
    path = dirname(dirname(__file__))
    meshfile = expanduser(join(path, 'data', 'periodic-hexagon.mesh'))

mesh = mfem.Mesh(meshfile, 1, 1)
dim = mesh.Dimension()

# 4. Define the ODE solver used for time integration. Several explicit
#    Runge-Kutta methods are available.
ode_solver = None
if ode_solver_type == 1:
    ode_solver = mfem.ForwardEulerSolver()
elif ode_solver_type == 2:
    ode_solver = mfem.RK2Solver(1.0)
elif ode_solver_type == 3:
    ode_solver = mfem.RK3SSolver()
elif ode_solver_type == 4:
    ode_solver = mfem.RK4Solver()
elif ode_solver_type == 6:
    ode_solver = mfem.RK6Solver()
else:
    print("Unknown ODE solver type: " + str(ode_solver_type))
    exit

# 5. Refine the mesh to increase the resolution. In this example we do
#    'ref_levels' of uniform refinement, where 'ref_levels' is a
#    command-line parameter. If the mesh is of NURBS type, we convert it to
#    a (piecewise-polynomial) high-order mesh.
for lev in range(ser_ref_levels):
    mesh.UniformRefinement()
    if mesh.NURBSext:
        mesh.SetCurvature(max(order, 1))
    bb_min, bb_max = mesh.GetBoundingBox(max(order, 1))


# 6. Define the parallel mesh by a partitioning of the serial mesh. Refine
#    this mesh further in parallel to increase the resolution. Once the
#    parallel mesh is defined, the serial mesh can be deleted.
pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
for k in range(par_ref_levels):
    pmesh.UniformRefinement()

# 7. Define the discontinuous DG finite element space of the given
#    polynomial order on the refined mesh.
fec = mfem.DG_FECollection(order, dim, mfem.BasisType.GaussLobatto)
fes = mfem.ParFiniteElementSpace(pmesh, fec)

global_vSize = fes.GlobalTrueVSize()
if myid == 0:
    print("Number of unknowns: " + str(global_vSize))

#
#  Define coefficient using VecotrPyCoefficient and PyCoefficient
#  A user needs to define EvalValue method
#


class velocity_coeff(mfem.VectorPyCoefficient):
    def EvalValue(self, x):
        dim = len(x)

        center = (bb_min + bb_max)/2.0
        # map to the reference [-1,1] domain
        X = 2 * (x - center) / (bb_max - bb_min)
        if problem == 0:
            if dim == 1:
                v = [1.0, ]
            elif dim == 2:
                v = [sqrt(2./3.), sqrt(1./3)]
            elif dim == 3:
                v = [sqrt(3./6.), sqrt(2./6), sqrt(1./6.)]
        elif (problem == 1 or problem == 2):
            # Clockwise rotation in 2D around the origin
            w = pi/2
            if dim == 1:
                v = [1.0, ]
            elif dim == 2:
                v = [w*X[1],  - w*X[0]]
            elif dim == 3:
                v = [w*X[1],  - w*X[0],  0]
        elif (problem == 3):
            # Clockwise twisting rotation in 2D around the origin
            w = pi/2
            d = max((X[0]+1.)*(1.-X[0]), 0.) * max((X[1]+1.)*(1.-X[1]), 0.)
            d = d ** 2
            if dim == 1:
                v = [1.0, ]
            elif dim == 2:
                v = [d*w*X[1],  - d*w*X[0]]
            elif dim == 3:
                v = [d*w*X[1],  - d*w*X[0],  0]
        return v


class u0_coeff(mfem.PyCoefficient):
    def EvalValue(self, x):
        dim = len(x)

        center = (bb_min + bb_max)/2.0
        # map to the reference [-1,1] domain
        X = 2 * (x - center) / (bb_max - bb_min)
        if (problem == 0 or problem == 1):
            if dim == 1:
                return exp(-40. * (X[0]-0.5)**2)
            elif (dim == 2 or dim == 3):
                rx = 0.45
                ry = 0.25
                cx = 0.
                cy = -0.2
                w = 10.
                if dim == 3:
                    s = (1. + 0.25*cos(2 * pi * x[2]))
                    rx = rx * s
                    ry = ry * s
                return (erfc(w * (X[0]-cx-rx)) * erfc(-w*(X[0]-cx+rx)) *
                        erfc(w * (X[1]-cy-ry)) * erfc(-w*(X[1]-cy+ry)))/16

        elif problem == 2:
            rho = hypot(x[0], x[1])
            phi = arctan2(x[1], x[0])
            return (sin(pi * rho) ** 2) * sin(3*phi)
        elif problem == 3:
            return sin(pi * X[0]) * sin(pi * X[1])

        return 0.0

# Inflow boundary condition (zero for the problems considered in this example)


class inflow_coeff(mfem.PyCoefficient):
    def EvalValue(self, x):
        return 0

# 8. Set up and assemble the bilinear and linear forms corresponding to the
#    DG discretization. The DGTraceIntegrator involves integrals over mesh
#    interior faces.


velocity = velocity_coeff(dim)
inflow = inflow_coeff()
u0 = u0_coeff()

m = mfem.ParBilinearForm(fes)
m.AddDomainIntegrator(mfem.MassIntegrator())
k = mfem.ParBilinearForm(fes)
k.AddDomainIntegrator(mfem.ConvectionIntegrator(velocity, -1.0))
k.AddInteriorFaceIntegrator(
    mfem.TransposeIntegrator(mfem.DGTraceIntegrator(velocity, 1.0, -0.5)))
k.AddBdrFaceIntegrator(
    mfem.TransposeIntegrator(mfem.DGTraceIntegrator(velocity, 1.0, -0.5)))

b = mfem.ParLinearForm(fes)
b.AddBdrFaceIntegrator(
    mfem.BoundaryFlowIntegrator(inflow, velocity, -1.0, -0.5))

m.Assemble()
m.Finalize()
skip_zeros = 0
k.Assemble(skip_zeros)
k.Finalize(skip_zeros)
b.Assemble()

M = m.ParallelAssemble()
K = k.ParallelAssemble()
B = b.ParallelAssemble()

# 7. Define the initial conditions, save the corresponding grid function to
#    a file
u = mfem.ParGridFunction(fes)
u.ProjectCoefficient(u0)
U = u.GetTrueDofs()

smyid = '{:0>6d}'.format(myid)
mesh_name = "ex9-mesh."+smyid
sol_name = "ex9-init."+smyid
pmesh.Print(mesh_name, 8)
u.Save(sol_name, 8)


class FE_Evolution(mfem.PyTimeDependentOperator):
    def __init__(self, M, K, b):
        mfem.PyTimeDependentOperator.__init__(self, M.Height())

        self.M_prec = mfem.HypreSmoother()
        self.M_solver = mfem.CGSolver(M.GetComm())
        self.z = mfem.Vector(M.Height())

        self.K = K
        self.M = M
        self.b = b
        self.M_prec.SetType(mfem.HypreSmoother.Jacobi)
        self.M_solver.SetPreconditioner(self.M_prec)
        self.M_solver.SetOperator(M)
        self.M_solver.iterative_mode = False
        self.M_solver.SetRelTol(1e-9)
        self.M_solver.SetAbsTol(0.0)
        self.M_solver.SetMaxIter(100)
        self.M_solver.SetPrintLevel(0)


#    def EvalMult(self, x):
#        if you want to impolement Mult in using python objects,
#        such as numpy.. this needs to be implemented and don't
#        overwrite Mult


    def Mult(self, x, y):
        self.K.Mult(x, self.z)
        self.z += b
        self.M_solver.Mult(self.z, y)


adv = FE_Evolution(M, K, B)

ode_solver.Init(adv)
t = 0.0
ti = 0
while True:
    if t > t_final - dt/2:
        break
    t, dt = ode_solver.Step(U, t, dt)
    ti = ti + 1
    if ti % vis_steps == 0:
        if myid == 0:
            print("time step: " + str(ti) + ", time: " + str(np.round(t, 3)))


u.Assign(U)
sol_name = "ex9-final."+smyid
u.Save(sol_name, 8)



# ----------------------------------------------------------------------------------------
# Global build parameters
# ----------------------------------------------------------------------------------------
import os

is_configured = False
prefix = ''

verbose = True
git_sshclone = False
swig_only = False
run_swig = False
clean_swig = False
build_mfem = False
mfem_branch = None
build_mfemp = False
build_metis = False
build_hypre = False
build_libceed = False
build_gslib = False
build_parallel = False
build_serial = False

ext_prefix = ''
mfem_outside = False
mfems_prefix = ''
mfemp_prefix = ''
mfem_source = os.path.join(os.path.dirname(__file__), "..", "external", "mfem")

metis_prefix = ''
hypre_prefix = ''

enable_cuda = False
enable_cuda_hypre = False
cuda_prefix = ''
cuda_arch = ''
enable_pumi = False
pumi_prefix = ''
enable_strumpack = False
strumpack_prefix = ''
enable_libceed = False
libceed_prefix = ''
libceed_only = False
enable_gslib = False
gslibs_prefix = ''
gslibp_prefix = ''
gslib_only = False
mfem_debug = False
mfem_build_miniapps = True

enable_suitesparse = False
suitesparse_prefix = "/usr/"

enable_lapack = False
blas_libraries = ""
lapack_libraries = ""

dry_run = False
do_bdist_wheel = False
bdist_wheel_dir = ''

keep_temp = False

use_unverifed_SSL = False if os.getenv(
    "unverifedSSL") is None else os.getenv("unverifiedSSL")

use_metis_gklib = False
metis_64 = False

# ----------------------------------------------------------------------------------------
#   command line configuration parameter (pip -C)
# ----------------------------------------------------------------------------------------

cfs = {}

# ----------------------------------------------------------------------------------------
#   enviromental variables.
# ----------------------------------------------------------------------------------------
cc_command = 'cc' if os.getenv("CC") is None else os.getenv("CC")
cxx_command = 'c++' if os.getenv("CC") is None else os.getenv("CXX")
mpicc_command = 'mpicc' if os.getenv("MPICC") is None else os.getenv("MPICC")
mpicxx_command = 'mpic++' if os.getenv(
    "MPICXX") is None else os.getenv("MPICXX")
cxxstd_flag = '-std=c++17' if os.getenv(
    "CXXSTDFLAG") is None else os.getenv("CXXSTDFLAG")

# location of MPI.h. Usually it is not needed, as long as MPI compiler can be used
#
mpiinc = '' if os.getenv("MPIINC") is None else os.getenv("MPIINC")



# ----------------------------------------------------------------------------------------
# Routines for metis
# ----------------------------------------------------------------------------------------
import sys
import os
import re
import subprocess
from sys import platform

__all__ = ["make_metis_gklib", "make_metis"]

from build_utils import *
from build_consts import *

import build_globals as bglb


def make_metis_gklib(use_int64=False, use_real64=False):
    '''
    build GKlib/metis
    '''

    '''
    build/install GKlib
    '''
    if bglb.verbose:
        print("Building gklib")

    path = os.path.join(extdir, 'gklib')
    if not bglb.dry_run and not os.path.exists(path):
        assert False, "gklib is not downloaded"

    path = os.path.join(path, 'cmbuild')
    if os.path.exists(path):
        print("working directory already exists!")
    else:
        os.makedirs(path)
    pwd = chdir(path)

    cmake_opts = {'DBUILD_SHARED_LIBS': '1',
                  'DCMAKE_INSTALL_PREFIX': metis_prefix}
    if verbose:
        cmake_opts['DCMAKE_VERBOSE_MAKEFILE'] = '1'

    cmake('..', **cmake_opts)
    make('gklib')
    make_install('gklib')
    os.chdir(pwd)

    '''
    build/install metis
    '''
    path = os.path.join(extdir, 'metis')
    if not bglb.dry_run and not os.path.exists(path):
        assert False, "metis is not downloaded"
    elif not os.path.exists(path):
        os.makedirs(path)
        os.makedirs(os.path.join(path, 'build'))

    pwd = chdir(path)

    gklibpath = os.path.dirname(find_libpath_from_prefix(
        'GKlib', bglb.metis_prefix))

    options = ['gklib_path='+bglb.metis_prefix]
    if use_int64:
        options.append('i64=1')

    if use_real64:
        options.append('r64=1')

    command = ['make', 'config', 'shared=1'] + options
    command = command + ['prefix=' + bglb.metis_prefix, 'cc=' + bglb.cc_command]
    make_call(command)

    chdir('build')
    cmake_opts = {'DGKLIB_PATH': bglb.metis_prefix,
                  'DSHARED': '1',
                  'DCMAKE_C_COMPILER': bglb.cc_command,
                  'DCMAKE_C_STANDARD_LIBRARIES': '-lGKlib',
                  'DCMAKE_INSTALL_RPATH': "@loader_path",
                  'DCMAKE_BUILD_WITH_INSTALL_RPATH': '1',
                  'DCMAKE_INSTALL_PREFIX': bglb.metis_prefix}
    if bglb.verbose:
        cmake_opts['DCMAKE_VERBOSE_MAKEFILE'] = '1'

    cmake('..', **cmake_opts)
    chdir(path)
    make('metis')
    make_install('metis')

    if platform == "darwin":
        command = ['install_name_tool',
                   '-id',
                   os.path.join("@rpath", 'libGKlib.dylib'),
#                   os.path.join(bglb.metis_prefix, 'lib', 'libGKlib.dylib'),
                   os.path.join(bglb.metis_prefix, 'lib', 'libGKlib.dylib'), ]
        make_call(command)
        command = ['install_name_tool',
                   '-id', 
                   os.path.join("@rpath", 'libmetis.dylib'),
                   os.path.join(bglb.metis_prefix, 'lib', 'libmetis.dylib'), ]
        make_call(command)
    os.chdir(pwd)


def make_metis(use_int64=False, use_real64=False):
    '''
    build metis
    '''
    if bglb.verbose:
        print("Building metis")

    path = os.path.join(extdir, 'metis')
    if not os.path.exists(path):
        assert False, "metis is not downloaded"

    pwd = chdir(path)

    if use_int64:
        pattern_int = "#define IDXTYPEWIDTH 32"
        replace_int = "#define IDXTYPEWIDTH 64"
    else:
        pattern_int = "#define IDXTYPEWIDTH 64"
        replace_int = "#define IDXTYPEWIDTH 32"
    with open("include/metis.h", "r") as metis_header_fid:
        metis_header_lines = metis_header_fid.readlines()
    with open("include/metis.h", "w") as metis_header_fid:
        for line in metis_header_lines:
            metis_header_fid.write(re.sub(pattern_int, replace_int, line))

    if use_real64:
        pattern_real = "#define REALTYPEWIDTH 32"
        replace_real = "#define REALTYPEWIDTH 64"
    else:
        pattern_real = "#define REALTYPEWIDTH 64"
        replace_real = "#define REALTYPEWIDTH 32"
    with open("include/metis.h", "r") as metis_header_fid:
        metis_header_lines = metis_header_fid.readlines()
    with open("include/metis.h", "w") as metis_header_fid:
        for line in metis_header_lines:
            metis_header_fid.write(re.sub(pattern_real, replace_real, line))

    command = ['make', 'config', 'shared=1',
               'prefix=' + bglb.metis_prefix,
               'cc=' + bglb.cc_command]
    make_call(command, env={'CMAKE_POLICY_VERSION_MINIMUM': '3.5'})
    make('metis')
    make_install('metis')

    if platform == "darwin":
        command = ['install_name_tool',
                   '-id',
                   os.path.join('@rpath', 'libmetis.dylib'),
                   os.path.join(bglb.metis_prefix, 'lib', 'libmetis.dylib'), ]
        make_call(command)
    os.chdir(pwd)
